# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

# Last Cleanup: 2026-01-04

from __future__ import annotations
import numpy as np
from dataclasses import dataclass, field
from typing import Literal

 
@dataclass
class Axis:
    """A representation of an axis.
    An Axis object always has length 1 and points in some 3D direction.
    By default XAX, YAX, and ZAX are constructed and defined in the global namespace of the
    FEM module.
    """
    vector: np.ndarray

    def __repr__(self) -> str:
        return f"Axis({self.vector})"
    
    def __post_init__(self):

        self.vector = self.vector/np.linalg.norm(self.vector)
        self.np: np.ndarray = self.vector
        self.x: float = self.vector[0]
        self.y: float = self.vector[1]
        self.z: float = self.vector[2]

    def __neg__(self):
        return Axis(-self.vector)
    
    @property
    def tuple(self) -> tuple[float, float, float]:
        return (self.x, self.y, self.z)
    
    @property
    def xy(self) -> tuple[float, float]:
        return (self.x, self.y)
    
    @property
    def neg(self) -> Axis:
        return Axis(-self.vector)
    
    def __iter__(self):
        yield self.x
        yield self.y
        yield self.z
        
    def cross(self, other: Axis) -> Axis:
        """Take the cross produt with another vector

        Args:
            other (Axis): Vector B in AxB

        Returns:
            Axis: The resultant Axis.
        """
        return Axis(np.cross(self.vector, other.vector))

    def dot(self, other: Axis) -> float:
        """Take the dot product of two vectors A·B

        Args:
            other (Axis): Vector B

        Returns:
            float: The resultant vector (A·B)
        """
        return self.x*other.x + self.y*other.y + self.z*other.z
    
    def pair(self, other: Axis) -> Plane:
        """Pair this vector with another to span the plane A⨂B

        Args:
            other (Axis): Vector B

        Returns:
            Plane: The plane spanned by A and B.
        """
        return Plane(self, other)
    
    def __mul__(self, other: Axis) -> Plane:
        """The multiply binary operator

        Args:
            other (Axis): _description_

        Returns:
            Plane: The output plane
        """
        return self.pair(other)
    
    def construct_cs(self, origin: tuple[float, float, float] = (0.,0.,0.)) -> CoordinateSystem:
        """Constructs a coordinate system where this vector is the Z-axis
        and the X and Y axis are normal to this axis but with an arbitrary rotation.

        Returns:
            CoordinateSystem: The resultant coordinate system
        """
        ax = Axis(np.array([1, 0, 0]))
        if np.abs(self.dot(ax)) > 0.999:
            ax = Axis(np.array([0, 1, 0]))
        ax1 = self.cross(ax)
        ax2 = self.cross(ax1).neg
        return CoordinateSystem(ax2, ax1, self, np.array(origin))


############################################################
#                         CONSTANTS                        #
############################################################

XAX: Axis = Axis(np.array([1, 0, 0]))
YAX: Axis = Axis(np.array([0, 1, 0]))
ZAX: Axis = Axis(np.array([0, 0, 1]))



############################################################
#                         FUNCTIONS                        #
############################################################

def _parse_vector(vec: np.ndarray | tuple[float, float, float] | list[float] | Axis | Frame) -> np.ndarray:
    """ Takes an array, tuple, list or Axis and alwasys returns an array."""
    if isinstance(vec, np.ndarray):
        return vec
    elif isinstance(vec, (list,tuple)):
        return np.array(vec)
    elif isinstance(vec, Axis):
        return vec.vector
    elif isinstance(vec, Frame):
        return vec.c0
    return np.array(vec)

def _parse_axis(vec: np.ndarray | tuple[float, float, float] | list[float] | Axis) -> Axis:
    """Takes an array, tuple, list or Axis and always returns an Axis.

    Args:
        vec (np.ndarray | tuple | list | Axis): The Axis data

    Returns:
        Axis: The Axis object.
    """
    if isinstance(vec, np.ndarray):
        return Axis(vec)
    elif isinstance(vec, (list,tuple)):
        return Axis(np.array(vec))
    elif isinstance(vec, Axis):
        return vec
    return Axis(np.array(vec))


def argparse_xyz(x: float | np.ndarray | tuple[float, float, float] | list[float] | Axis | Frame,
                 y: float | None = None,
                 z: float | None = None) -> tuple[float, float, float]:
    """A helper function to parse xyz arguments that can be given
    as single arrays, tuples, lists, Axis or Point objects or as separate
    x,y,z float values.
    Args:
        x (float | np.ndarray | tuple | list | Axis | Point): The x value or entire coordinate
        y (float | None, optional): The y value. Defaults to None.
        z (float | None, optional): The z value. Defaults to None.
    """
    if y is None and z is None:
        if isinstance(x, (float, int, np.floating, np.integer)):
            raise ValueError('If x is a single float value, y and z must also be provided.')
        return tuple(_parse_vector(x))
    else:
        if y is None:
            y = 0.0
        if z is None:
            z = 0.0
        return (x, y, z)
    
############################################################
#                        PLANE CLASS                       #
############################################################

@dataclass
class Plane:
    """A generalization of any plane of inifinite size spanned by two Axis objects.

    """
    uax: Axis
    vax: Axis

    def __post_init__(self):
        # Check if axes are orthogonal
        if not np.isclose(np.dot(self.uax.vector, self.vax.vector), 0):
            raise ValueError("Axes are not orthogonal")

    def __repr__(self) -> str:
        return f"Plane({self.uax}, {self.vax})"
    
    def __mul__(self, other: Axis) -> CoordinateSystem:
        return CoordinateSystem(self.uax, self.vax, other)
    
    @property
    def normal(self) -> Axis:
        """Returns the normal of the plane as u ⨉ v.

        Returns:
            Axis: The axis object normal to the plane.
        """
        return self.uax.cross(self.vax)
    
    def flip(self) -> Plane:
        """Flips the planes U and V axes.

        Returns:
            Plane: A new plane object.
        """
        return Plane(self.vax, self.uax)
    
    def cs(self, x0: float = 0, y0: float = 0, z0: float = 0) -> CoordinateSystem:
        """Returns a CoordinateSystem object for the plane where the XY axes are aligned
        with the plane UV axis and Z is normal.

        Args:
            x0 (float): The x coordinate of the origin
            y0 (float): The y coordinate of the origin
            z0 (float): The z coordinate of the origin

        Returns:
            CoordinateSystem: The coordinate system object
        """
        origin = np.array([x0, y0, z0])
        return CoordinateSystem(self.uax, self.vax, self.normal, origin)
    
    def grid(self, 
             uax: np.ndarray | tuple[float, float, int], 
             vax: np.ndarray | tuple[float, float, int], 
             origin: np.ndarray | tuple[float, float, float],
             indexing: Literal['xy','ij'] = 'xy') -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Spans a grid of points in the plane based on a np.linspace like argument type.
        The first uax argument should be a start, finish, Npoints type tuple of a float, float and integer.
        Item for the vax. The origin defines the coordinate at which u,v = 0 will be placed.
        The return type is an N,M np.meshgrid defined by the indexing 'xy' or 'ij'.

        Args:
            uax (np.ndarray | tuple[float, float, int]): The uax linspace argument values
            vax (np.ndarray | tuple[float, float, int]): The vax linspace argument values
            origin (np.ndarray | tuple[float, float, float]): The origin for u,v = 0
            indexing (Literal[&#39;xy&#39;,&#39;ij&#39;], optional): The indexing type. Defaults to 'xy'.

        Returns:
            tuple[np.ndarray, np.ndarray, np.ndarray]: The X, Y, Z (N,M) meshgrid of coordinates.
        """
        if isinstance(uax, tuple):
            uax = np.linspace(*uax)
        if isinstance(vax, tuple):
            vax = np.linspace(*vax)
        if isinstance(origin, tuple):
            origin = np.array(origin)

        U, V = np.meshgrid(uax, vax, indexing=indexing)
        uax = U.flatten()
        vax = V.flatten()
        shp = U.shape
        xs = self.uax.np[0]*uax + self.vax.np[0]*vax + origin[0]
        ys = self.uax.np[1]*uax + self.vax.np[1]*vax + origin[1]
        zs = self.uax.np[2]*uax + self.vax.np[2]*vax + origin[2]

        return xs.reshape(shp), ys.reshape(shp), zs.reshape(shp)
    
    def span(self, u: float, v: float, N: int, origin: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Create a grid of XYZ coordinates in the plane reaching from 0 to u and 0 to v in N steps at the given origin

        Args:
            u (float): The distance along the first axis
            v (float): The distance along the second axis 
            N (int): The number of sample points per axis
            origin (np.ndarray): The origin of the planar grid.

        Returns:
            tuple[np.ndarray, np.ndarray, np.ndarray]: The set of X,Y,Z coordinates.
        """
        uax = np.linspace(0, u, N)
        vax = np.linspace(0, v, N)
        U, V = np.meshgrid(uax, vax, indexing='ij')
        uax = U.flatten()
        vax = V.flatten()
        shp = U.shape
        xs = self.uax.np[0]*uax + self.vax.np[0]*vax + origin[0]
        ys = self.uax.np[1]*uax + self.vax.np[1]*vax + origin[1]
        zs = self.uax.np[2]*uax + self.vax.np[2]*vax + origin[2]

        return xs.reshape(shp), ys.reshape(shp), zs.reshape(shp)


############################################################
#                         CONSTANTS                        #
############################################################

XYPLANE = Plane(XAX, YAX)
XZPLANE = Plane(XAX, ZAX)
YZPLANE = Plane(YAX, ZAX)
YXPLANE = Plane(YAX, XAX)
ZXPLANE = Plane(ZAX, XAX)
ZYPLANE = Plane(ZAX, YAX)

############################################################
#                  COORDINATE SYSTEM CLASS                 #
############################################################


@dataclass
class CoordinateSystem:
    """A class representing CoordinateSystem information.

    This class is widely used throughout the FEM solver to embed objects in space properly.
    The x,y and z unit vectors are defined by Axis objects. The origin by a np.ndarray.

    The property _is_global is should only be set for any CoordinateSystem class that is wished to
    be considered as global. This is reserved for the GCS intance create automatically with:
        xax = (1,0,0)
        yax = (0,1,0)
        zax = (0,0,1)
        origin = (0., 0., 0.) meters
    """
    xax: Axis
    yax: Axis
    zax: Axis
    origin: np.ndarray = field(default_factory=lambda: np.array([0,0,0]))
    _basis: np.ndarray = field(default=None)
    _basis_inv: np.ndarray = field(default=None)
    _is_global: bool = False

    def __post_init__(self):
        self.xax = _parse_axis(self.xax)
        self.yax = _parse_axis(self.yax)
        self.zax = _parse_axis(self.zax)
        self._basis = np.array([self.xax.np, self.yax.np, self.zax.np]).T
        self._basis_inv = np.linalg.pinv(self._basis)

    def __repr__(self) -> str:
        return f"CS({self.xax}, {self.yax}, {self.zax}, {self.origin})"
    
    def copy(self) -> CoordinateSystem:
        """ Creates a copy of this coordinate system."""
        return CoordinateSystem(self.xax, self.yax, self.zax, self.origin)
    
    def displace(self, dx: float, dy: float, dz: float) -> CoordinateSystem:
        """Creates a displaced version of this coordinate system. The basis is kept the same.

        Args:
            dx (float): The X-displacement (meters)
            dy (float): The Y-displacement (meters)
            dz (float): The Z-displacement (meters)

        Returns:
            CoordinateSystem: The new CoordinateSystem object.
        """
        csnew = CoordinateSystem(self.xax, self.yax, self.zax, self.origin + np.array([dx, dy, dz]))
        return csnew
    
    def rotate(self, 
               axis: tuple | list | np.ndarray | Axis, 
               angle: float, 
               degrees: bool = True,
               origin: bool | np.ndarray = False) -> CoordinateSystem:
        """Return a new CoordinateSystem rotated about the given axis (through the global origin)
        by `angle`. If `degrees` is True, `angle` is interpreted in degrees.

        Args:
            axis (tuple | list | np.ndarray | Axis): The rotation axis
            angle (float): The rotation angle (in degrees if degrees = True)
            degrees (bool, optional): Whether to use degrees. Defaults to True.
            origin (bool, np.array, optional): Whether to rotate the origin as well. Defaults to False.

        Returns:
            CoordinateSystem: The new rotated coordinate system
        """

        # Convert to radians if needed
        if degrees:
            theta = angle * np.pi/180

        # Normalize the rotation axis
        u = _parse_vector(axis)
        u = u / np.linalg.norm(u)

        # Build the skew-symmetric cross-product matrix K for u
        K = np.array([
            [   0,   -u[2],  u[1]],
            [ u[2],     0,  -u[0]],
            [-u[1],  u[0],     0]
        ], dtype=np.float64)

        # Rodrigues' rotation formula: R = I + sinθ·K + (1−cosθ)·K²
        Imat = np.eye(3, dtype=np.float64)
        R = Imat + np.sin(theta) * K + (1 - np.cos(theta)) * (K @ K)

        # Rotate each axis and the origin
        new_x = R @ self.xax.vector
        new_y = R @ self.yax.vector
        new_z = R @ self.zax.vector
        
        if origin is not False:
            if isinstance(origin, bool):
                new_o = R @ self.origin
            else:
                new_o = (R @ (self.origin-np.array(origin))) + np.array(origin)
        else:
            new_o = self.origin.copy()

        return CoordinateSystem(
            xax=new_x,
            yax=new_y,
            zax=new_z,
            origin=new_o,
            _is_global=False
        )
    
    def swapxy(self) -> None:
        """Swaps the XY axes of the CoordinateSystem.
        """
        self.xax, self.yax = self.yax, self.xax
        self.__post_init__()
    
    def affine_from_global(self) -> np.ndarray:
        """Returns an Affine transformation matrix in order to transform coordinates from
        the global coordinate system to this coordinate system.

        Returns:
            np.ndarray: The affine transformation matrix.
        """
        x = self.xax.vector
        y = self.yax.vector
        z = self.zax.vector
        o = self.origin
        T = np.eye(4, dtype=np.float64)
        T[0:3, 0] = x
        T[0:3, 1] = y
        T[0:3, 2] = z
        T[0:3, 3] = o
        return T
    
    def affine_to_global(self) -> np.ndarray:
        """Returns an Affine transformation matrix in order to transform coordinates from
        this local coordinate system to the coordinate system.

        Returns:
            np.ndarray: The affine transformation matrix.
        """
        T = self.affine_from_global()
        R = T[0:3, 0:3]
        o = T[0:3, 3]
        R_inv = np.linalg.inv(R)
        o_new = - R_inv @ o
        T_inv = np.eye(4, dtype=np.float64)
        T_inv[0:3, 0:3] = R_inv
        T_inv[0:3, 3]   = o_new
        return T_inv
    
    def in_global_cs(self, x: np.ndarray, y: np.ndarray, z: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Converts x,y,z coordinates into the global coordinate system.

        Args:
            x (np.ndarray): The x-coordinates (meter)
            y (np.ndarray): The y-coordinates (meter)
            z (np.ndarray): The z-coordinates (meter)

        Returns:
            tuple[np.ndarray, np.ndarray, np.ndarray]: The resultant x, y and z coordinates.
        """
        xg = self.xax.np[0]*x + self.yax.np[0]*y + self.zax.np[0]*z + self.origin[0]
        yg = self.xax.np[1]*x + self.yax.np[1]*y + self.zax.np[1]*z + self.origin[1]
        zg = self.xax.np[2]*x + self.yax.np[2]*y + self.zax.np[2]*z + self.origin[2]
        return xg, yg, zg
    
    def in_local_cs(self, 
                    x: np.ndarray,
                    y: np.ndarray,
                    z: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Converts x,y,z coordinates into the local coordinate system.

        Args:
            x (np.ndarray): The x-coordinates (meter)
            y (np.ndarray): The y-coordinates (meter)
            z (np.ndarray): The z-coordinates (meter)

        Returns:
            tuple[np.ndarray, np.ndarray, np.ndarray]: The resultant x, y and z coordinates.
        """
        B = self._basis_inv
        xg = x - self.origin[0]
        yg = y - self.origin[1]
        zg = z - self.origin[2]
        x = B[0,0]*xg + B[0,1]*yg + B[0,2]*zg
        y = B[1,0]*xg + B[1,1]*yg + B[1,2]*zg
        z = B[2,0]*xg + B[2,1]*yg + B[2,2]*zg
        return x, y, z
    
    def in_global_basis(self, 
                        x: np.ndarray,
                        y: np.ndarray,
                        z: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Converts x,y,z vector components into the global coordinate basis.

        Args:
            x (np.ndarray): The x-vector components (meter)
            y (np.ndarray): The y-vector components (meter)
            z (np.ndarray): The z-vector components (meter)

        Returns:
            tuple[np.ndarray, np.ndarray, np.ndarray]: The resultant x, y and z vectors.
        """
        xg = self.xax.np[0]*x + self.yax.np[0]*y + self.zax.np[0]*z
        yg = self.xax.np[1]*x + self.yax.np[1]*y + self.zax.np[1]*z
        zg = self.xax.np[2]*x + self.yax.np[2]*y + self.zax.np[2]*z
        return xg, yg, zg
    
    def in_local_basis(self, 
                          x: np.ndarray,
                          y: np.ndarray,
                          z: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Converts x,y,z vector components into the local coordinate basis.

        Args:
            x (np.ndarray): The x-vector components (meter)
            y (np.ndarray): The y-vector components (meter)
            z (np.ndarray): The z-vector components (meter)

        Returns:
            tuple[np.ndarray, np.ndarray, np.ndarray]: The resultant x, y and z vectors.
        """
        B = self._basis_inv
        xg = x
        yg = y
        zg = z
        x = B[0,0]*xg + B[0,1]*yg + B[0,2]*zg
        y = B[1,0]*xg + B[1,1]*yg + B[1,2]*zg
        z = B[2,0]*xg + B[2,1]*yg + B[2,2]*zg
        return x, y, z
    
    @property
    def gx(self) -> float:
        """ The origin x-coordinate in global coordinates."""
        return self.origin[0]
    
    @property
    def gy(self) -> float:
        """ The origin y-coordinate in global coordinates."""
        return self.origin[1]
    
    @property
    def gz(self) -> float:
        """ The origin z-coordinate in global coordinates."""
        return self.origin[2]
    
    @property
    def gxhat(self) -> np.ndarray:
        """ The x-axis unit vector in global coordinates."""
        return self.xax.np
    
    @property
    def gyhat(self) -> np.ndarray:
        """ The y-axis unit vector in global coordinates."""
        return self.yax.np
    
    @property
    def gzhat(self) -> np.ndarray:
        """ The z-axis unit vector in global coordinates."""
        return self.zax.np


############################################################
#                         CONSTANTS                        #
############################################################

CS = CoordinateSystem
GCS = CoordinateSystem(XAX, YAX, ZAX, np.zeros(3), _is_global=True)


############################################################
#                         FUNCTIONS                        #
############################################################

def cs(axes: str = 'xyz', origin: tuple[float, float, float] = (0.,0.,0.,)) -> CoordinateSystem:
    """Generate a coordinate system based on a simple string
    The string must contain the letters x, X, y, Y, z and/or Z. 
    Small letters refer to positive axes and capitals to negative axes.
    
    The string must be 3 characters long.
    
    The first position indices which global axis gets assigned to the new local X-axis
    The second position indicates the Y-axis
    The third position indicates the Z-axis
    
    Thus, rotating the global XYZ coordinate system 90 degrees around the Z axis would yield: yXz

    Args:
        axes (str): The axis description
        origin (tuple[float, float, float], optional): The origin of the coordinate system. Defaults to (0.,0.,0.,).

    Returns:
        CoordinateSystem: The resultant coordinate system
    """
    if len(axes) != 3:
        raise ValueError('Axis object must be of length 3')
    
    axlib = {
        'x': Axis(np.array([1.0,0.,0.])),
        'X': Axis(np.array([-1.0,0.,0.])),
        'y': Axis(np.array([0.,1.0,0.])),
        'Y': Axis(np.array([0.,-1.0,0.])),
        'z': Axis(np.array([0.,0.,1.0])),
        'Z': Axis(np.array([0.,0.,-1.0])),
    }
    ax_obj = [axlib[ax] for ax in axes]
    
    return (ax_obj[0]*ax_obj[1]*ax_obj[2]).displace(*origin)



@dataclass
class Frame:
    """A Frame is a generalization of a coordinate plus a local 3D axis system
    
    Frames behave like coordinates when passed as arguments to functions
    that require coordinates. Additionally, they can be used in the .stick()
    method to move anchors of objects aligned with other ancors.
    

    Raises:
        AttributeError: _description_
        ValueError: _description_

    Returns:
        _type_: _description_

    Yields:
        _type_: _description_
    """
    c0: np.ndarray
    _x: np.ndarray = field(default_factory=lambda: np.array([1.0, 0.0, 0.0]))
    _y: np.ndarray = field(default_factory=lambda: np.array([0.0, 1.0, 0.0]))
    _z: np.ndarray = field(default_factory=lambda: np.array([0.0, 0.0, 1.0]))

    def __post_init__(self):
        self.c0 = np.array(self.c0)
        self._x = np.array(self._x)
        self._y = np.array(self._y)
        self._z = np.array(self._z)
    
    def __getattr__(self, name: str) -> Frame:
        """ A frame with the coordinate systems xyz axis oriented
        
        in the provided order.
        Lower case means negative and upper-case positive.
        Example:
         >>> frame.yZx is the axis system (-y, +z, -x)"""
        if len(name) != 3:
            raise AttributeError(f'There is no attribute named: {name}')
        axes = {
            'x': -self._x,
            'X': self._x,
            'y': -self._y,
            'Y': self._y,
            'z': -self._z,
            'Z': self._z
        }
        return Frame(self.c0, *[axes[c] for c in name])
    
    def __add__(self, other: Frame) -> Frame:
        """ Adds two frames by adding their origins and keeping the axis of self.

        Args:
            other (Frame): The other frame to add
        """
        return Frame(self.c0 + other.c0, self._x, self._y, self._z)
    
    @property
    def tx(self) -> Frame:
        """ Same frame rotated around its positive X-axis 180 degrees"""
        return Frame(self.c0, self._x, -self._y, -self._z)
    
    @property
    def ty(self) -> Frame:
        """ Same frame rotated around its positive Y-axis 180 degrees"""
        return Frame(self.c0, -self._x, self._y, -self._z)
    
    @property
    def tz(self) -> Frame:
        """ Same frame rotated around its positive Z-axis 180 degrees"""
        return Frame(self.c0, -self._x, -self._y, self._z)
    
    @property
    def mx(self) -> Frame:
        """ Same frame mirrored in its X-axis"""
        return Frame(self.c0, -self._x, self._y, self._z)
    
    @property
    def my(self) -> Frame:
        """ Same frame mirrored in its Y-axis"""
        return Frame(self.c0, self._x, -self._y, self._z)
    
    @property
    def mz(self) -> Frame:
        """ Same frame mirrored in its Z-axis"""
        return Frame(self.c0, self._x, self._y, -self._z)
    
    def __iter__(self):
        yield self.c0[0]
        yield self.c0[1]
        yield self.c0[2]
    
    def as_homogeneous(self) -> np.ndarray:
        R = np.column_stack((self._x, self._y, self._z))
        T = np.eye(4)
        T[:3, :3] = R
        T[:3, 3] = self.c0
        return T
    
    def cs(self) -> CoordinateSystem:
        """ Creates a CoordinateSystem object at this frame with the given axis as Z-axis.

        Args:
            axis (tuple | np.ndarray | Axis): The Z-axis direction

        Returns:
            CoordinateSystem: The resultant CoordinateSystem object.
        """
        return CoordinateSystem(self._x, self._y, self._z, self.c0)
    
    def compute_affine(self, other: Frame) -> np.ndarray:
        """
        Compute affine transform mapping this frame to `other`.
        Returns a 4x4 homogeneous transformation matrix.
        """
        T_self = self.as_homogeneous()
        T_other = other.as_homogeneous()
        print(T_self)
        print(T_other)
        return T_other @ np.linalg.inv(T_self)

    def rotate(self, c0, ax, angle):
        """
        Rotate self.point and self.axis about axis `ax`, centered at `c0`, by `angle` radians.

        Parameters
        ----------
        c0 : np.ndarray
            The center of rotation, shape (3,).
        ax : np.ndarray
            The axis to rotate around, shape (3,). Need not be unit length.
        angle : float
            Rotation angle in radians.
        """
        
        angle = -angle
        # Ensure axis is a unit vector
        k = ax / np.linalg.norm(ax)

        # Precompute trig values
        cos_theta = np.cos(angle)
        sin_theta = np.sin(angle)

        def rodrigues(v: np.ndarray) -> np.ndarray:
            """
            Rotate vector v around axis k by angle using Rodrigues' formula.
            """
            # term1 = v * cosθ
            term1 = v * cos_theta
            # term2 = (k × v) * sinθ
            term2 = np.cross(k, v) * sin_theta
            # term3 = k * (k ⋅ v) * (1 - cosθ)
            term3 = k * (np.dot(k, v)) * (1 - cos_theta)
            return term1 + term2 + term3

        # Rotate the origin point about c0:
        rel_o = self.c0 - c0            # move to rotation-centre coordinates
        rot_o = rodrigues(rel_o)       # rotate
        self.c0 = rot_o + c0            # move back

        # Rotate the normal vector (pure direction, no translation)
        self._x = rodrigues(self._x)
        self._y = rodrigues(self._y)
        self._z = rodrigues(self._z)

    def translate(self, dx: float, dy: float, dz: float):
        """_summary_

        Args:
            dx (_type_): _description_
            dy (_type_): _description_
            dz (_type_): _description_
        """
        self.c0 = self.c0 + np.array([dx, dy, dz])
    
    def mirror(self, c0: np.ndarray, pln: np.ndarray) -> None:
        """
        Reflect self.o and self.n across the plane passing through c0
        with normal pln.

        Parameters
        ----------
        c0 : np.ndarray
            A point on the mirror plane, shape (3,).
        pln : np.ndarray
            The normal of the mirror plane, shape (3,). Need not be unit length.
        """
        # Normalize the plane normal
        k = pln / np.linalg.norm(pln)
        
        # Reflect the origin point:
        # compute vector from plane point to self.o
        v_o = self.c0 - c0
        # signed distance along normal
        dist_o = np.dot(v_o, k)
        # reflection
        self.c0 = self.c0 - 2 * dist_o * k

        # Reflect the normal/direction vector:
        dist_x = np.dot(self._x, k)
        dist_y = np.dot(self._y, k)
        dist_z = np.dot(self._z, k)
        self._x = (self._x - 2 * dist_x * k)
        self._y = (self._y - 2 * dist_y * k)
        self._z = (self.Z - 2 * dist_z * k)

    def affine_transform(self, M: np.ndarray):
        """
        Apply a 4×4 affine transformation matrix to both self.o and self.n.

        Parameters
        ----------
        M : np.ndarray
            The 4×4 affine transformation matrix.
            - When applied to a point, use homogeneous w=1.
            - When applied to a direction/vector, use homogeneous w=0.
        """
        # Validate shape
        if M.shape != (4, 4):
            raise ValueError(f"Expected M to be 4×4, got shape {M.shape}")

        # Transform origin point (homogeneous w=1)
        homo_o = np.empty(4)
        homo_o[:3] = self.c0
        homo_o[3] = 1.0
        transformed_o = M @ homo_o
        
        self.c0 = transformed_o[:3]

        # Transform normal/direction vector (homogeneous w=0)
        homo_x = np.empty(4)
        homo_x[:3] = self._x
        homo_x[3] = 0.0
        
        homo_y = np.empty(4)
        homo_y[:3] = self._y
        homo_y[3] = 0.0
        
        homo_z = np.empty(4)
        homo_z[:3] = self._z
        homo_z[3] = 0.0
        
        self._x = (M @ homo_x)[:3]
        self._y = (M @ homo_y)[:3]
        self._z = (M @ homo_z)[:3]
        



from mumps_emerge import ZMumpsContext
from scipy import sparse
import numpy as np

# Reuse factorizations by running `job=3` with new right hand sides
# or analyses by supplying new values and running `job=2` to repeat
# the factorization process.

class MUMPSInterface:
    def __init__(self):
        self.ctx = ZMumpsContext(1,0)
        self.ctx.set_silent()
        self._analysed = False
        self._factorized = False
        # Parallel analysis
        #self.ctx.id.icntl[27] = 1   # sequential analysis
        #self.ctx.id.icntl[47] = 1
        self.ctx.id.icntl[6]  = 5   # METIS
        self.ctx.id.icntl[27] = 1      # ICNTL(28)
        self.ctx.id.icntl[28] = 0      # ICNTL(29) automatic (safe)
        self.ctx.id.icntl[47] = 0
        
    def analyse_matrix(self, A: sparse.csr_matrix):
        # Convert to COO
        A = A.tocoo()
        n = A.shape[0]

        if self.ctx.myid == 0:
            # 1-based indices
            irn = A.row.astype(np.int32) + 1
            jcn = A.col.astype(np.int32) + 1
            a = A.data.astype(np.complex128)

            self.ctx.set_shape(n)
            self.ctx.set_centralized_assembled(irn, jcn, a)

        # Analysis
        self.ctx.run(job=1)
        self._analysed = True
        self._factorized = False

    def factorize(self, A: sparse.csr_matrix):
        if not self._analysed:
            raise RuntimeError("Matrix must be analysed before factorization")
        A = A.tocoo()

        n = A.shape[0]

        if self.ctx.myid == 0:
            # 1-based indices
            irn = A.row.astype(np.int32) + 1
            jcn = A.col.astype(np.int32) + 1
            a = A.data.astype(np.complex128)

            self.ctx.set_shape(n)
            self.ctx.set_centralized_assembled(irn, jcn, a)
            
        self.ctx.run(job=2)
        self._factorized = True

    def solve(self, b):
        if not self._factorized:
            raise RuntimeError("Matrix must be factorized before solve")
        if self.ctx.myid == 0:
            x = np.asarray(b, dtype=np.complex128).copy()
            self.ctx.set_rhs(x)
        else:
            x = None

        self.ctx.run(job=3)

        return x, 0

    def destroy(self):
        self.ctx.destroy()
        


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from .microwave_data import MWField
import numpy as np
from ...mth.optimized import matmul, outward_normal
from numba import njit, f8, c16, i8, types, prange, b1# type: ignore, p
from loguru import logger
from ...const import C0, MU0, EPS0


############################################################
#                    OPTIMIZED FUNCTIONS                   #
############################################################

@njit(types.Tuple((f8[:,:], f8[:]))(f8[:,:], i8[:,:], f8[:], b1[:]), nogil=True, cache=True)
def tet_to_node(nodes, tets, sizes, included):
    """
    Parameters
    ----------
    nodes : (3, N) float64
        Node coordinates.
    tets : (4, M) int64
        Node indices for each tetrahedron.
    sizes : (M,) float64
        Requested mesh sizes per tetrahedron.
    included : (M,) boolean
        Whether a tetrahedron imposes a size constraint.

    Returns
    -------
    coords_out : (K, 3) float64
        Coordinates of nodes that received a constraint.
    sizes_out : (K,) float64
        Minimum size constraint per returned node.
    """
    N = nodes.shape[1]
    M = tets.shape[1]

    big = 1e300
    node_sizes = np.full(N, big)
    constrained = np.zeros(N, dtype=np.uint8)  # 0/1 flag

    # Accumulate min size per node from included tets
    for e in range(M):
        if included[e]:
            s = sizes[e]
            # four vertices per tet
            n0 = tets[0, e]
            n1 = tets[1, e]
            n2 = tets[2, e]
            n3 = tets[3, e]

            if constrained[n0] == 0:
                node_sizes[n0] = s
                constrained[n0] = 1
            else:
                if s < node_sizes[n0]:
                    node_sizes[n0] = s

            if constrained[n1] == 0:
                node_sizes[n1] = s
                constrained[n1] = 1
            else:
                if s < node_sizes[n1]:
                    node_sizes[n1] = s

            if constrained[n2] == 0:
                node_sizes[n2] = s
                constrained[n2] = 1
            else:
                if s < node_sizes[n2]:
                    node_sizes[n2] = s

            if constrained[n3] == 0:
                node_sizes[n3] = s
                constrained[n3] = 1
            else:
                if s < node_sizes[n3]:
                    node_sizes[n3] = s

    # Count constrained nodes
    K = 0
    for i in range(N):
        if constrained[i] == 1:
            K += 1

    # Pack outputs
    coords_out = np.empty((3, K), dtype=nodes.dtype)
    sizes_out = np.empty(K, dtype=sizes.dtype)

    j = 0
    for i in range(N):
        if constrained[i] == 1:
            coords_out[0, j] = nodes[0, i]
            coords_out[1, j] = nodes[1, i]
            coords_out[2, j] = nodes[2, i]
            sizes_out[j] = node_sizes[i]
            j += 1

    return coords_out, sizes_out

@njit(cache=True, nogil=True)
def diam_circum_circle(v1: np.ndarray, v2: np.ndarray, v3: np.ndarray, eps: float = 1e-14) -> float:
    """
    Diameter of the circumcircle of triangle (v1,v2,v3) in 3D.
    Uses D = (|AB|*|AC|*|BC|) / ||AB x AC||.
    Returns np.inf if points are (near) collinear.
    """
    AB = v2 - v1
    AC = v3 - v1
    BC = v3 - v2

    a = np.linalg.norm(BC)
    b = np.linalg.norm(AC)
    c = np.linalg.norm(AB)

    cross = np.cross(AB, AC)
    denom = np.linalg.norm(cross)  # 2 * area of triangle

    if denom < eps:
        return np.inf  # degenerate/collinear

    return (a * b * c) / denom  # diameter = 2R

@njit(cache=True, nogil=True)
def circum_sphere_diam(v1: np.ndarray, v2: np.ndarray, v3: np.ndarray, v4: np.ndarray, eps: float = 1e-14) -> float:
    """
    Diameter of the circumsphere of tetrahedron (v1,v2,v3,v4) in 3D.
    Solves for center c from 2*(pi - p4) · c = |pi|^2 - |p4|^2, i=1..3.
    Returns np.inf if points are (near) coplanar/degenerate.
    """
    p1, p2, p3, p4 = v1, v2, v3, v4

    M = np.empty((3, 3), dtype=np.float64)
    M[0, :] = 2.0 * (p1 - p4)
    M[1, :] = 2.0 * (p2 - p4)
    M[2, :] = 2.0 * (p3 - p4)

    # manual 3x3 determinant (Numba-friendly)
    det = (
        M[0,0] * (M[1,1]*M[2,2] - M[1,2]*M[2,1])
        - M[0,1] * (M[1,0]*M[2,2] - M[1,2]*M[2,0])
        + M[0,2] * (M[1,0]*M[2,1] - M[1,1]*M[2,0])
    )
    if np.abs(det) < eps:
        return np.inf  # coplanar/degenerate

    rhs = np.empty(3, dtype=np.float64)
    rhs[0] = np.sum(p1**2) - np.sum(p4**2)
    rhs[1] = np.sum(p2**2) - np.sum(p4**2)
    rhs[2] = np.sum(p3**2) - np.sum(p4**2)

    # Solve for circumcenter
    c = np.linalg.solve(M, rhs)

    # Radius = distance to any vertex
    R = np.linalg.norm(c - p1)
    return 2.0 * R  # diameter

def print_sparam_matrix(pre: str, S: np.ndarray):
    """
    Print an N x N complex S-parameter matrix in dB∠deg format.
    Magnitude in dB rounded to 2 decimals, phase in degrees with 1 decimal.
    """
    S = np.asarray(S)
    if S.ndim != 2 or S.shape[0] != S.shape[1]:
        raise ValueError("S must be a square (N x N) complex matrix")

    N = S.shape[0]
    logger.debug(pre+"S-parameter matrix (dB ∠ deg):")

    for i in range(N):
        row_str = []
        for j in range(N):
            mag_db = 20 * np.log10(np.abs(S[i, j]) + np.finfo(float).eps)
            phase_deg = np.degrees(np.angle(S[i, j]))
            row_str.append(f"{mag_db:6.2f} dB ∠ {phase_deg:6.1f}°")
        logger.debug(" | ".join(row_str))

def compute_convergence(Sold: np.ndarray, Snew: np.ndarray) -> float:
    """
    Return a single scalar: max |Snew - Sold|.
    Works for shapes (N,N) or (..., N, N); reduces over all axes.
    """
    
    Sold = np.asarray(Sold)
    Snew = np.asarray(Snew)
    print_sparam_matrix('Old:',Sold)
    print_sparam_matrix('New',Snew)
    if Sold.shape != Snew.shape:
        raise ValueError("Sold and Snew must have identical shapes")
    #amp_conv = float(np.abs(np.abs(Snew) - np.abs(Sold)).max())
    mag_conv = float(np.abs(np.abs(Snew)-np.abs(Sold)).max())
    amp_conv = float(np.abs(Snew - Sold).max())
    phase_conv = float(np.abs(np.angle(np.diag(Sold)/np.diag(Snew))).max()) * 180/np.pi
    return amp_conv, mag_conv, phase_conv

def select_refinement_indices(errors: np.ndarray, refine: float) -> np.ndarray:
    """
    Dörfler marking:
    Choose the minimal number of elements whose squared error sum
    reaches at least 'refine' (theta in [0,1]) of the global squared error.

    Args:
        errors: 1D or ND array of local error indicators (nonnegative).
        refine: theta in [0,1]; fraction of total error energy to target.

    Returns:
        np.ndarray of indices (ints) sorted by decreasing error.
    """
    # Flatten and sanitize
    errs = np.abs(errors)
    
    ind = errs * errs
    total = np.sum(ind)
    
    # Sort by decreasing indicator
    order = np.argsort(ind)[::-1]
    sum_error = 0
    indices = []
    for index in order:
        sum_error += ind[index]
        indices.append(index)
        if sum_error >= refine*total and len(indices) >= 20:
            break
    
    return np.array(indices)

@njit(f8[:](i8, f8[:,:], f8, f8, f8[:]), cache=True, nogil=True, parallel=False)
def compute_size(index: int, coords: np.ndarray, gr: float, scaler: float, dss: np.ndarray) -> float:
    """Optimized function to compute the size impressed by size constraint points on each other size constraint point.

    Args:
        id (int): _description_
        coords (np.ndarray): _description_
        gr (float): growth rate
        scaler (float): _description_
        dss (np.ndarray): _description_

    Returns:
        float: _description_
    """
    N = dss.shape[0]
    sizes = np.zeros((N,), dtype=np.float64)-1.0
    x, y, z = coords[:,index]
    for n in prange(N):
        if n == index:
            sizes[n] = dss[n]*scaler
            continue
        distance = ((x-coords[0,n])**2 + (y-coords[1,n])**2 + (z-coords[2,n])**2)**0.5 
        nsize = scaler*dss[n]/gr - (1-gr)/gr * max(0, (distance - dss[n]*0))
        sizes[n] = nsize
    return sizes

@njit(f8[:](f8[:,:], i8, i8[:]), cache=True, nogil=True, parallel=True)
def nbmin(matrix, axis, include):
    
    if axis==0:
        N = matrix.shape[1]
        out = np.empty((N,), dtype=np.float64)
        for n in prange(N):
            out[n] = np.min(matrix[include==1,n])
        return out
    if axis==1:
        N = matrix.shape[0]
        out = np.empty((N,), dtype=np.float64)
        for n in prange(N):
            out[n] = np.min(matrix[n,include==1])
        return out
    else:
        out = np.empty((N,), dtype=np.float64)
        return out   
    
@njit(i8(i8, f8[:,:], f8, i8[:]), cache=True, nogil=True, parallel=False)
def can_remove(index: int, M: np.ndarray, scaling: float, include: np.ndarray) -> int:

    ratio = np.min(M[index,:] / nbmin(M, 1, include))
    
    if ratio > scaling:
        return 1
    return 0

@njit(i8[:](f8[:,:], f8, f8[:], f8, f8), cache=True, nogil=True, parallel=False)
def reduce_point_set(coords: np.ndarray, gr: float, dss: np.ndarray, scaler: float, keep_percentage: float) -> list[int]:
    N = dss.shape[0]
    impressed_size = np.zeros((N,N), np.float64)
    
    include = np.ones((N,), dtype=np.int64)
    
    for n in range(N):
        impressed_size[n,:] = compute_size(n, coords, gr, scaler, dss)
    
    
    current_min = nbmin(impressed_size, 1, include)
    counter = 0
    
    for i in range(N):

        if (N-counter)/N < keep_percentage:
            break
        
        if (N-counter) <= 10:
            break
        
        if current_min[i] <= impressed_size[i,i]*0.8:
            include[i] = 0
            counter = counter + 1
            current_min = nbmin(impressed_size, 1, include)
        
        
    
    ids = np.arange(N)
    output = ids[include==1]
    return output


@njit(i8[:, :](i8[:], i8[:, :]), cache=True, nogil=True)
def local_mapping(vertex_ids, triangle_ids):
    """
    Parameters
    ----------
    vertex_ids   : 1-D int64 array (length 4)
        Global vertex 0.1005964238ers of one tetrahedron, in *its* order
        (v0, v1, v2, v3).

    triangle_ids : 2-D int64 array (nTri × 3)
        Each row is a global-ID triple of one face that belongs to this tet.

    Returns
    -------
    local_tris   : 2-D int64 array (nTri × 3)
        Same triangles, but every entry replaced by the local index
        0,1,2,3 that the vertex has inside this tetrahedron.
        (Guaranteed to be ∈{0,1,2,3}; no -1 ever appears if the input
        really belongs to the tet.)
    """
    ndim = triangle_ids.shape[0]
    ntri = triangle_ids.shape[1]
    out  = np.zeros(triangle_ids.shape, dtype=np.int64)

    for t in range(ntri):                 # each triangle
        for j in range(ndim):                # each vertex in that triangle
            gid = triangle_ids[j, t]      # global ID to look up

            # linear search over the four tet vertices
            for k in range(4):
                if vertex_ids[k] == gid:
                    out[j, t] = k         # store local index 0-3
                    break                 # stop the k-loop

    return out

@njit(f8(f8[:], f8[:], f8[:]), cache = True, nogil=True)
def compute_volume(xs, ys, zs):
    x1, x2, x3, x4 = xs
    y1, y2, y3, y4 = ys
    z1, z2, z3, z4 = zs

    return np.abs(-x1*y2*z3/6 + x1*y2*z4/6 + x1*y3*z2/6 - x1*y3*z4/6 - x1*y4*z2/6 + x1*y4*z3/6 + x2*y1*z3/6 - x2*y1*z4/6 - x2*y3*z1/6 + x2*y3*z4/6 + x2*y4*z1/6 - x2*y4*z3/6 - x3*y1*z2/6 + x3*y1*z4/6 + x3*y2*z1/6 - x3*y2*z4/6 - x3*y4*z1/6 + x3*y4*z2/6 + x4*y1*z2/6 - x4*y1*z3/6 - x4*y2*z1/6 + x4*y2*z3/6 + x4*y3*z1/6 - x4*y3*z2/6)

@njit(types.Tuple((f8[:], f8[:], f8[:], f8[:], f8))(f8[:], f8[:], f8[:]), cache = True, nogil=True)
def tet_coefficients(xs, ys, zs):
    ## THIS FUNCTION WORKS
    x1, x2, x3, x4 = xs
    y1, y2, y3, y4 = ys
    z1, z2, z3, z4 = zs

    aas = np.empty((4,), dtype=np.float64)
    bbs = np.empty((4,), dtype=np.float64)
    ccs = np.empty((4,), dtype=np.float64)
    dds = np.empty((4,), dtype=np.float64)

    V = np.abs(-x1*y2*z3/6 + x1*y2*z4/6 + x1*y3*z2/6 - x1*y3*z4/6 - x1*y4*z2/6 + x1*y4*z3/6 + x2*y1*z3/6 - x2*y1*z4/6 - x2*y3*z1/6 + x2*y3*z4/6 + x2*y4*z1/6 - x2*y4*z3/6 - x3*y1*z2/6 + x3*y1*z4/6 + x3*y2*z1/6 - x3*y2*z4/6 - x3*y4*z1/6 + x3*y4*z2/6 + x4*y1*z2/6 - x4*y1*z3/6 - x4*y2*z1/6 + x4*y2*z3/6 + x4*y3*z1/6 - x4*y3*z2/6)
    
    aas[0] = x2*y3*z4 - x2*y4*z3 - x3*y2*z4 + x3*y4*z2 + x4*y2*z3 - x4*y3*z2
    aas[1] = -x1*y3*z4 + x1*y4*z3 + x3*y1*z4 - x3*y4*z1 - x4*y1*z3 + x4*y3*z1
    aas[2] = x1*y2*z4 - x1*y4*z2 - x2*y1*z4 + x2*y4*z1 + x4*y1*z2 - x4*y2*z1
    aas[3] = -x1*y2*z3 + x1*y3*z2 + x2*y1*z3 - x2*y3*z1 - x3*y1*z2 + x3*y2*z1
    bbs[0] = -y2*z3 + y2*z4 + y3*z2 - y3*z4 - y4*z2 + y4*z3
    bbs[1] = y1*z3 - y1*z4 - y3*z1 + y3*z4 + y4*z1 - y4*z3
    bbs[2] = -y1*z2 + y1*z4 + y2*z1 - y2*z4 - y4*z1 + y4*z2
    bbs[3] = y1*z2 - y1*z3 - y2*z1 + y2*z3 + y3*z1 - y3*z2
    ccs[0] = x2*z3 - x2*z4 - x3*z2 + x3*z4 + x4*z2 - x4*z3
    ccs[1] = -x1*z3 + x1*z4 + x3*z1 - x3*z4 - x4*z1 + x4*z3
    ccs[2] = x1*z2 - x1*z4 - x2*z1 + x2*z4 + x4*z1 - x4*z2
    ccs[3] = -x1*z2 + x1*z3 + x2*z1 - x2*z3 - x3*z1 + x3*z2
    dds[0] = -x2*y3 + x2*y4 + x3*y2 - x3*y4 - x4*y2 + x4*y3
    dds[1] = x1*y3 - x1*y4 - x3*y1 + x3*y4 + x4*y1 - x4*y3
    dds[2] = -x1*y2 + x1*y4 + x2*y1 - x2*y4 - x4*y1 + x4*y2
    dds[3] = x1*y2 - x1*y3 - x2*y1 + x2*y3 + x3*y1 - x3*y2

    return aas, bbs, ccs, dds, V

DPTS_2D = np.array([[0.22338159, 0.22338159, 0.22338159, 0.10995174, 0.10995174,
        0.10995174],
       [0.10810302, 0.44594849, 0.44594849, 0.81684757, 0.09157621,
        0.09157621],
       [0.44594849, 0.44594849, 0.10810302, 0.09157621, 0.09157621,
        0.81684757],
       [0.44594849, 0.10810302, 0.44594849, 0.09157621, 0.81684757,
        0.09157621]], dtype=np.float64)

DPTS_3D = np.array([[-0.078933  ,  0.04573333,  0.04573333,  0.04573333,  0.04573333,
         0.14933333,  0.14933333,  0.14933333,  0.14933333,  0.14933333,
         0.14933333],
       [ 0.25      ,  0.78571429,  0.07142857,  0.07142857,  0.07142857,
         0.39940358,  0.39940358,  0.39940358,  0.10059642,  0.10059642,
         0.10059642],
       [ 0.25      ,  0.07142857,  0.07142857,  0.07142857,  0.78571429,
         0.10059642,  0.10059642,  0.39940358,  0.39940358,  0.39940358,
         0.10059642],
       [ 0.25      ,  0.07142857,  0.07142857,  0.78571429,  0.07142857,
         0.39940358,  0.10059642,  0.10059642,  0.39940358,  0.10059642,
         0.39940358],
       [ 0.25      ,  0.07142857,  0.78571429,  0.07142857,  0.07142857,
         0.10059642,  0.39940358,  0.10059642,  0.10059642,  0.39940358,
         0.39940358]], dtype=np.float64)

@njit(c16[:,:](f8[:,:], f8[:,:], c16[:], i8[:,:], i8[:,:]), cache=True, nogil=True)
def compute_field(coords: np.ndarray, 
                 vertices: np.ndarray,
                 Etet: np.ndarray, 
                 l_edge_ids: np.ndarray, 
                 l_tri_ids: np.ndarray):
    
    x = coords[0,:]
    y = coords[1,:]
    z = coords[2,:]
    N = coords.shape[1]
    
    xvs = vertices[0,:]
    yvs = vertices[1,:]
    zvs = vertices[2,:]
    
    a_s, b_s, c_s, d_s, V = tet_coefficients(xvs, yvs, zvs)
    
    Em1s = Etet[0:6]
    Ef1s = Etet[6:10]
    Em2s = Etet[10:16]
    Ef2s = Etet[16:20]
    
    Exl = np.zeros_like(x, dtype=np.complex128)
    Eyl = np.zeros_like(x, dtype=np.complex128)
    Ezl = np.zeros_like(x, dtype=np.complex128)
    
    V1 = (6*V)**3
    
    for ie in range(6):
        Em1, Em2 = Em1s[ie], Em2s[ie]
        edgeids = l_edge_ids[:, ie]
        a1, a2 = a_s[edgeids]
        b1, b2 = b_s[edgeids]
        c1, c2 = c_s[edgeids]
        d1, d2 = d_s[edgeids]
        x1, x2 = xvs[edgeids]
        y1, y2 = yvs[edgeids]
        z1, z2 = zvs[edgeids]
        F1 = (a1 + b1*x + c1*y + d1*z)
        F2 = (a2 + b2*x + c2*y + d2*z)
        L = np.sqrt((x1 - x2)**2 + (y1 - y2)**2 + (z1 - z2)**2)
        ex =  L*(Em1*F1 + Em2*F2)*(b1*F2 - b2*F1)/V1
        ey =  L*(Em1*F1 + Em2*F2)*(c1*F2 - c2*F1)/V1
        ez =  L*(Em1*F1 + Em2*F2)*(d1*F2 - d2*F1)/V1

        Exl += ex
        Eyl += ey
        Ezl += ez
    
    for ie in range(4):
        Em1, Em2 = Ef1s[ie], Ef2s[ie]
        triids = l_tri_ids[:, ie]
        a1, a2, a3 = a_s[triids]
        b1, b2, b3 = b_s[triids]
        c1, c2, c3 = c_s[triids]
        d1, d2, d3 = d_s[triids]

        x1, x2, x3 = xvs[l_tri_ids[:, ie]]
        y1, y2, y3 = yvs[l_tri_ids[:, ie]]
        z1, z2, z3 = zvs[l_tri_ids[:, ie]]

        L1 = np.sqrt((x1-x3)**2 + (y1-y3)**2 + (z1-z3)**2)
        L2 = np.sqrt((x1-x2)**2 + (y1-y2)**2 + (z1-z2)**2)
        
        F1 = (a1 + b1*x + c1*y + d1*z)
        F2 = (a2 + b2*x + c2*y + d2*z)
        F3 = (a3 + b3*x + c3*y + d3*z)
        
        Q1 = Em1*L1*F2
        Q2 = Em2*L2*F3
        ex =  (-Q1*(b1*F3 - b3*F1) + Q2*(b1*F2 - b2*F1))/V1
        ey =  (-Q1*(c1*F3 - c3*F1) + Q2*(c1*F2 - c2*F1))/V1
        ez =  (-Q1*(d1*F3 - d3*F1) + Q2*(d1*F2 - d2*F1))/V1
        
        Exl += ex
        Eyl += ey
        Ezl += ez

    out = np.zeros((3,N), dtype=np.complex128)
    out[0,:] = Exl
    out[1,:] = Eyl
    out[2,:] = Ezl
    return out

@njit(c16[:,:](f8[:,:], f8[:,:], c16[:], i8[:,:], i8[:,:]), cache=True, nogil=True)
def compute_curl(coords: np.ndarray, 
                 vertices: np.ndarray,
                 Etet: np.ndarray, 
                 l_edge_ids: np.ndarray, 
                 l_tri_ids: np.ndarray):
    
    x = coords[0,:]
    y = coords[1,:]
    z = coords[2,:]
    
    xvs = vertices[0,:]
    yvs = vertices[1,:]
    zvs = vertices[2,:]
    
    a_s, b_s, c_s, d_s, V = tet_coefficients(xvs, yvs, zvs)
    
    Em1s = Etet[0:6]
    Ef1s = Etet[6:10]
    Em2s = Etet[10:16]
    Ef2s = Etet[16:20]
    
    Exl = np.zeros((x.shape[0],), dtype=np.complex128)
    Eyl = np.zeros((x.shape[0],), dtype=np.complex128)
    Ezl = np.zeros((x.shape[0],), dtype=np.complex128)
    
    V1 = 216*V**3
    V2 = 72*V**3
    
    for ie in range(6):
        Em1, Em2 = Em1s[ie], Em2s[ie]
        edgeids = l_edge_ids[:, ie]
        a1, a2 = a_s[edgeids]
        b1, b2 = b_s[edgeids]
        c1, c2 = c_s[edgeids]
        d1, d2 = d_s[edgeids]
        x1, x2 = xvs[edgeids]
        y1, y2 = yvs[edgeids]
        z1, z2 = zvs[edgeids]

        L = np.sqrt((x1 - x2)**2 + (y1 - y2)**2 + (z1 - z2)**2)
        C1 = Em1*a1
        C2 = Em1*b1
        C3 = Em1*c1
        C4 = Em1*c2
        C5 = Em2*a2
        C6 = Em2*b2
        C7 = Em2*c1
        C8 = Em2*c2
        C9 = Em1*b2
        C10 = Em2*b1
        D1 = c1*d2
        D2 = c2*d1
        D3 = d1*d2
        D4 = d1*d1
        D5 = c2*d2
        D6 = d2*d2
        D7 = b1*d2
        D8 = b2*d1
        D9 = c1*d1
        D10 = b2*d2
        D11 = b1*c2
        D12 = b2*c1
        D13 = c1*c2
        D14 = c1*c1
        D15 = b2*c2
        
        ex =  L*(-C1*D1 + C1*D2 - C2*D1*x + C2*D2*x - C3*D1*y + C3*D2*y - C3*D3*z + C4*D4*z - C5*D1 + C5*D2 - C6*D1*x + C6*D2*x - C7*D5*y - C7*D6*z + C8*D2*y + C8*D3*z)/V2
        ey =  L*(C1*D7 - C1*D8 + C2*D7*x - C2*D8*x + C2*D1*y + C2*D3*z - C9*D9*y - C9*D4*z + C5*D7 - C5*D8 + C10*D10*x + C10*D5*y + C10*D6*z - C6*D8*x - C6*D2*y - C6*D3*z)/V2
        ez =  L*(-C1*D11 + C1*D12 - C2*D11*x + C2*D12*x - C2*D13*y - C2*D2*z + C9*D14*y + C9*D9*z - C5*D11 + C5*D12 - C10*D15*x - C10*c2*c2*y - C10*D5*z + C6*D12*x + C6*D13*y + C6*D1*z)/V2
        Exl += ex
        Eyl += ey
        Ezl += ez
    
    for ie in range(4):
        Em1, Em2 = Ef1s[ie], Ef2s[ie]
        triids = l_tri_ids[:, ie]
        a1, a2, a3 = a_s[triids]
        b1, b2, b3 = b_s[triids]
        c1, c2, c3 = c_s[triids]
        d1, d2, d3 = d_s[triids]

        x1, x2, x3 = xvs[l_tri_ids[:, ie]]
        y1, y2, y3 = yvs[l_tri_ids[:, ie]]
        z1, z2, z3 = zvs[l_tri_ids[:, ie]]

        L1 = np.sqrt((x1-x3)**2 + (y1-y3)**2 + (z1-z3)**2)
        L2 = np.sqrt((x1-x2)**2 + (y1-y2)**2 + (z1-z2)**2)
        F1 = (a3 + b3*x + c3*y + d3*z)
        F2 = (a1 + b1*x + c1*y + d1*z)
        F3 = (a2 + b2*x + c2*y + d2*z)
        N1 = (d1*F1 - d3*F2)
        N2 = (c1*F1 - c3*F2)
        N3 = (c1*d3 - c3*d1)
        N4 = (d1*F3 - d2*F2)
        N5 = (c1*F3 - c2*F2)
        D1 = c1*d2
        D2 = c2*d1
        N6 = (D1 - D2)
        N7 = (b1*F1 - b3*F2)
        N8 = (b1*d3 - b3*d1)
        N9 = (b1*F3 - b2*F2)
        D7 = b1*d2
        D8 = b2*d1
        N10 = (D7 - D8)
        D11 = b1*c2
        D12 = b2*c1
        ex =  (Em1*L1*(-c2*N1 + d2*N2 + 2*N3*F3) - Em2*L2*(-c3*N4 + d3*N5 + 2*N6*F1))/V1
        ey =  (-Em1*L1*(-b2*N1 + d2*N7 + 2*N8*F3) + Em2*L2*(-b3*N4 + d3*N9 + 2*N10*F1))/V1
        ez =  (Em1*L1*(-b2*N2 + c2*N7 + 2*(b1*c3 - b3*c1)*F3) - Em2*L2*(-b3*N5 + c3*N9 + 2*(D11 - D12)*F1))/V1
        
        Exl += ex
        Eyl += ey
        Ezl += ez

    out = np.zeros((3,x.shape[0]), dtype=np.complex128)
    out[0,:] = Exl
    out[1,:] = Eyl
    out[2,:] = Ezl
    return out

@njit(c16[:](f8[:,:], f8[:,:], c16[:], i8[:,:], i8[:,:], c16[:,:]), cache=True, nogil=True)
def compute_div(coords: np.ndarray, 
                 vertices: np.ndarray,
                 Etet: np.ndarray, 
                 l_edge_ids: np.ndarray, 
                 l_tri_ids: np.ndarray,
                 Um: np.ndarray):
    
    uxx, uxy, uxz = Um[0,0], Um[0,1], Um[0,2]
    uyx, uyy, uyz = Um[1,0], Um[1,1], Um[1,2]
    uzx, uzy, uzz = Um[2,0], Um[2,1], Um[2,2]
    
    xs = coords[0,:]
    ys = coords[1,:]
    zs = coords[2,:]
    
    xvs = vertices[0,:]
    yvs = vertices[1,:]
    zvs = vertices[2,:]
    
    a_s, b_s, c_s, d_s, V = tet_coefficients(xvs, yvs, zvs)
    
    Em1s = Etet[0:6]
    Ef1s = Etet[6:10]
    Em2s = Etet[10:16]
    Ef2s = Etet[16:20]
    
    difE = np.zeros((xs.shape[0],), dtype=np.complex128)
    
    V1 = (216*V**3)
    V2 = (72*V**3)
    
    for ie in range(6):
        Em1, Em2 = Em1s[ie], Em2s[ie]
        edgeids = l_edge_ids[:, ie]
        a1, a2 = a_s[edgeids]
        b1, b2 = b_s[edgeids]
        c1, c2 = c_s[edgeids]
        d1, d2 = d_s[edgeids]
        x1, x2 = xvs[edgeids]
        y1, y2 = yvs[edgeids]
        z1, z2 = zvs[edgeids]

        L = np.sqrt((x1 - x2)**2 + (y1 - y2)**2 + (z1 - z2)**2)
        C1 = (a2 + b2*xs + c2*ys + d2*zs)
        C2 = (a1 + b1*xs + c1*ys + d1*zs)
        C3 = (b1*C1 - b2*C2)
        C4 = (c1*C1 - c2*C2)
        C5 = (d1*C1 - d2*C2)
        C6 = (b1*c2 - b2*c1)
        C7 = (c1*d2 - c2*d1)
        C8 = (b1*d2 - b2*d1)
        difE += (Em1*L*(b1*uxx*C3 + b1*uxy*C4 + b1*uxz*C5 + c1*uyx*C3 + c1*uyy*C4 + c1*uyz*C5 + d1*uzx*C3 
                       + d1*uzy*C4 + d1*uzz*C5 - uxy*C6*C2 - uxz*C8*C2 + uyx*C6*C2 - uyz*C7*C2 + uzx*C8*C2 + uzy*C7*C2) + 
                Em2*L*(b2*uxx*C3 + b2*uxy*C4 + b2*uxz*C5 + c2*uyx*C3 + c2*uyy*C4 + c2*uyz*C5 + d2*uzx*C3 
                       + d2*uzy*C4 + d2*uzz*C5 - uxy*C6*C1 - uxz*C8*C1 + uyx*C6*C1 - uyz*C7*C1 + uzx*C8*C1 + uzy*C7*C1))/V1
    
    for ie in range(4):
        Em1, Em2 = Ef1s[ie], Ef2s[ie]
        triids = l_tri_ids[:, ie]
        a1, a2, a3 = a_s[triids]
        b1, b2, b3 = b_s[triids]
        c1, c2, c3 = c_s[triids]
        d1, d2, d3 = d_s[triids]

        x1, x2, x3 = xvs[l_tri_ids[:, ie]]
        y1, y2, y3 = yvs[l_tri_ids[:, ie]]
        z1, z2, z3 = zvs[l_tri_ids[:, ie]]

        L1 = np.sqrt((x1-x3)**2 + (y1-y3)**2 + (z1-z3)**2)
        L2 = np.sqrt((x1-x2)**2 + (y1-y2)**2 + (z1-z2)**2)
        C1 = (a3 + b3*xs + c3*ys + d3*zs)
        C2 = (a1 + b1*xs + c1*ys + d1*zs)
        C6 = (a2 + b2*xs + c2*ys + d2*zs)
        C3 = (b1*C1 - b3*C2)
        C4 = (c1*C1 - c3*C2)
        C5 = (d1*C1 - d3*C2)
        C7 = (b1*c3 - b3*c1)
        C8 = (b1*d3 - b3*d1)
        C9 = (c1*d3 - c3*d1)
        C10 = (b1*C6 - b2*C2)
        C11 = (c1*C6 - c2*C2)
        C12 = (d1*C6 - d2*C2)
        C13 = (b1*c2 - b2*c1)
        C14 = (b1*d2 - b2*d1)
        C15 = (c1*d2 - c2*d1)
        
        difE += (-Em1*L1*(b2*uxx*C3 + b2*uxy*C4 + b2*uxz*C5 + c2*uyx*C3 + c2*uyy*C4 + c2*uyz*C5 + d2*uzx*C3 
                         + d2*uzy*C4 + d2*uzz*C5 - uxy*C7*C6 - uxz*C8*C6 + uyx*C7*C6 - uyz*C9*C6 + uzx*C8*C6 + uzy*C9*C6) 
                + Em2*L2*(b3*uxx*C10 + b3*uxy*C11 + b3*uxz*C12 + c3*uyx*C10 + c3*uyy*C11 + c3*uyz*C12 
                          + d3*uzx*C10 + d3*uzy*C11 + d3*uzz*C12 - uxy*C13*C1 - uxz*C14*C1 + uyx*C13*C1 - uyz*C15*C1 + uzx*C14*C1 + uzy*C15*C1))/V1
    
    return difE

@njit(c16[:](f8[:,:], c16[:], i8[:,:], i8[:,:], c16[:,:]), cache=True, nogil=True)
def compute_curl_curl(
                    vertices: np.ndarray,
                    Etet: np.ndarray, 
                    l_edge_ids: np.ndarray, 
                    l_tri_ids: np.ndarray,
                    Um: np.ndarray):
    
    uxx, uxy, uxz = Um[0,0], Um[0,1], Um[0,2]
    uyx, uyy, uyz = Um[1,0], Um[1,1], Um[1,2]
    uzx, uzy, uzz = Um[2,0], Um[2,1], Um[2,2]
    
    xvs = vertices[0,:]
    yvs = vertices[1,:]
    zvs = vertices[2,:]
    
    Exl = 0.0 + 0.0j
    Eyl = 0.0 + 0.0j
    Ezl = 0.0 + 0.0j
    
    a_s, b_s, c_s, d_s, V = tet_coefficients(xvs, yvs, zvs)
    
    Em1s = Etet[0:6]
    Ef1s = Etet[6:10]
    Em2s = Etet[10:16]
    Ef2s = Etet[16:20]
    
    V1 = (6*V)**3
    
    for ie in range(6):
        Em1, Em2 = Em1s[ie], Em2s[ie]
        edgeids = l_edge_ids[:, ie]
        a1, a2 = a_s[edgeids]
        b1, b2 = b_s[edgeids]
        c1, c2 = c_s[edgeids]
        d1, d2 = d_s[edgeids]
        x1, x2 = xvs[edgeids]
        y1, y2 = yvs[edgeids]
        z1, z2 = zvs[edgeids]

        L1 = np.sqrt((x1 - x2)**2 + (y1 - y2)**2 + (z1 - z2)**2)
        ex = -3*L1*(Em1*(b1*c1*c2*uzz - b1*c1*d2*uzy - b1*c2*d1*uyz + b1*d1*d2*uyy - b2*c1**2*uzz + b2*c1*d1*uyz + b2*c1*d1*uzy - b2*d1**2*uyy + c1**2*d2*uzx - c1*c2*d1*uzx - c1*d1*d2*uyx + c2*d1**2*uyx) + Em2*(b1*c2**2*uzz - b1*c2*d2*uyz - b1*c2*d2*uzy + b1*d2**2*uyy - b2*c1*c2*uzz + b2*c1*d2*uyz + b2*c2*d1*uzy - b2*d1*d2*uyy + c1*c2*d2*uzx - c1*d2**2*uyx - c2**2*d1*uzx + c2*d1*d2*uyx))
        ey = 3*L1*(Em1*(b1**2*c2*uzz - b1**2*d2*uzy - b1*b2*c1*uzz + b1*b2*d1*uzy + b1*c1*d2*uzx - b1*c2*d1*uxz - b1*c2*d1*uzx + b1*d1*d2*uxy + b2*c1*d1*uxz - b2*d1**2*uxy - c1*d1*d2*uxx + c2*d1**2*uxx) + Em2*(b1*b2*c2*uzz - b1*b2*d2*uzy - b1*c2*d2*uxz + b1*d2**2*uxy - b2**2*c1*uzz + b2**2*d1*uzy + b2*c1*d2*uxz + b2*c1*d2*uzx - b2*c2*d1*uzx - b2*d1*d2*uxy - c1*d2**2*uxx + c2*d1*d2*uxx))
        ez = -3*L1*(Em1*(b1**2*c2*uyz - b1**2*d2*uyy - b1*b2*c1*uyz + b1*b2*d1*uyy - b1*c1*c2*uxz + b1*c1*d2*uxy + b1*c1*d2*uyx - b1*c2*d1*uyx + b2*c1**2*uxz - b2*c1*d1*uxy - c1**2*d2*uxx + c1*c2*d1*uxx) + Em2*(b1*b2*c2*uyz - b1*b2*d2*uyy - b1*c2**2*uxz + b1*c2*d2*uxy - b2**2*c1*uyz + b2**2*d1*uyy + b2*c1*c2*uxz + b2*c1*d2*uyx - b2*c2*d1*uxy - b2*c2*d1*uyx - c1*c2*d2*uxx + c2**2*d1*uxx))
        
        Exl += ex/V1
        Eyl += ey/V1
        Ezl += ez/V1
    
    for ie in range(4):
        Em1, Em2 = Ef1s[ie], Ef2s[ie]
        triids = l_tri_ids[:, ie]
        a1, a2, a3 = a_s[triids]
        b1, b2, b3 = b_s[triids]
        c1, c2, c3 = c_s[triids]
        d1, d2, d3 = d_s[triids]

        x1, x2, x3 = xvs[l_tri_ids[:, ie]]
        y1, y2, y3 = yvs[l_tri_ids[:, ie]]
        z1, z2, z3 = zvs[l_tri_ids[:, ie]]

        L1 = np.sqrt((x1-x3)**2 + (y1-y3)**2 + (z1-z3)**2)
        L2 = np.sqrt((x1-x2)**2 + (y1-y2)**2 + (z1-z2)**2)
        
        ex = (Em1*L1*(3*c2*uzx*(c1*d3 - c3*d1) + 3*c2*uzz*(b1*c3 - b3*c1) - 3*d2*uyx*(c1*d3 - c3*d1) + 3*d2*uyy*(b1*d3 - b3*d1) - uyz*(-b2*(c1*d3 - c3*d1) + c2*(b1*d3 - b3*d1) + 2*d2*(b1*c3 - b3*c1)) - uzy*(b2*(c1*d3 - c3*d1) + 2*c2*(b1*d3 - b3*d1) + d2*(b1*c3 - b3*c1))) - Em2*L2*(3*c3*uzx*(c1*d2 - c2*d1) + 3*c3*uzz*(b1*c2 - b2*c1) - 3*d3*uyx*(c1*d2 - c2*d1) + 3*d3*uyy*(b1*d2 - b2*d1) - uyz*(-b3*(c1*d2 - c2*d1) + c3*(b1*d2 - b2*d1) + 2*d3*(b1*c2 - b2*c1)) - uzy*(b3*(c1*d2 - c2*d1) + 2*c3*(b1*d2 - b2*d1) + d3*(b1*c2 - b2*c1))))
        ey = (Em1*L1*(3*b2*uzy*(b1*d3 - b3*d1) - 3*b2*uzz*(b1*c3 - b3*c1) + 3*d2*uxx*(c1*d3 - c3*d1) - 3*d2*uxy*(b1*d3 - b3*d1) + uxz*(-b2*(c1*d3 - c3*d1) + c2*(b1*d3 - b3*d1) + 2*d2*(b1*c3 - b3*c1)) - uzx*(2*b2*(c1*d3 - c3*d1) + c2*(b1*d3 - b3*d1) - d2*(b1*c3 - b3*c1))) - Em2*L2*(3*b3*uzy*(b1*d2 - b2*d1) - 3*b3*uzz*(b1*c2 - b2*c1) + 3*d3*uxx*(c1*d2 - c2*d1) - 3*d3*uxy*(b1*d2 - b2*d1) + uxz*(-b3*(c1*d2 - c2*d1) + c3*(b1*d2 - b2*d1) + 2*d3*(b1*c2 - b2*c1)) - uzx*(2*b3*(c1*d2 - c2*d1) + c3*(b1*d2 - b2*d1) - d3*(b1*c2 - b2*c1))))
        ez = -(Em1*L1*(3*b2*uyy*(b1*d3 - b3*d1) - 3*b2*uyz*(b1*c3 - b3*c1) + 3*c2*uxx*(c1*d3 - c3*d1) + 3*c2*uxz*(b1*c3 - b3*c1) - uxy*(b2*(c1*d3 - c3*d1) + 2*c2*(b1*d3 - b3*d1) + d2*(b1*c3 - b3*c1)) - uyx*(2*b2*(c1*d3 - c3*d1) + c2*(b1*d3 - b3*d1) - d2*(b1*c3 - b3*c1))) - Em2*L2*(3*b3*uyy*(b1*d2 - b2*d1) - 3*b3*uyz*(b1*c2 - b2*c1) + 3*c3*uxx*(c1*d2 - c2*d1) + 3*c3*uxz*(b1*c2 - b2*c1) - uxy*(b3*(c1*d2 - c2*d1) + 2*c3*(b1*d2 - b2*d1) + d3*(b1*c2 - b2*c1)) - uyx*(2*b3*(c1*d2 - c2*d1) + c3*(b1*d2 - b2*d1) - d3*(b1*c2 - b2*c1)))) 
        
        Exl += ex/V1
        Eyl += ey/V1
        Ezl += ez/V1

    out = np.zeros((3,), dtype=np.complex128)
    out[0] = Exl
    out[1] = Eyl
    out[2] = Ezl
    return out

@njit(c16[:,:](c16[:], c16[:,:]), cache=True, fastmath=True, nogil=True)
def cross_c_arry(a: np.ndarray, b: np.ndarray):
    """Optimized complex single vector cross product

    Args:
        a (np.ndarray): (3,) vector a
        b (np.ndarray): (3,) vector b

    Returns:
        np.ndarray: a ⨉ b
    """
    crossv = np.empty((3,b.shape[1]), dtype=np.complex128)
    crossv[0,:] = a[1]*b[2,:] - a[2]*b[1,:]
    crossv[1,:] = a[2]*b[0,:] - a[0]*b[2,:]
    crossv[2,:] = a[0]*b[1,:] - a[1]*b[0,:]
    return crossv

@njit(c16[:](c16[:], c16[:,:]), cache=True, fastmath=True, nogil=True)
def dot_c_arry(a: np.ndarray, b: np.ndarray):
    """Optimized complex single vector cross product

    Args:
        a (np.ndarray): (3,) vector a
        b (np.ndarray): (3,) vector b

    Returns:
        np.ndarray: a ⨉ b
    """
    dotv = a[0]*b[0,:] + a[1]*b[1,:] + a[2]*b[2,:]
    return dotv

@njit(types.Tuple((f8[:], f8[:]))(f8[:,:], i8[:,:], i8[:,:], i8[:,:], f8[:,:],
                                  c16[:], f8[:], f8[:], i8[:,:], i8[:,:],
                                  f8[:,:], i8[:,:], i8[:,:], c16[:], c16[:], i8[:], f8), cache=True, nogil=True)
def compute_error_single(nodes, tets, tris, edges, centers, 
                         Efield, 
                         edge_lengths,
                         areas,
                         tet_to_edge, 
                         tet_to_tri, 
                         tri_centers, 
                         tri_to_tet,
                         tet_to_field, 
                         er, 
                         ur,
                         pec_tris,
                         k0,) -> np.ndarray:

    tet_is_pec = np.zeros((tris.shape[1],), dtype=np.bool_)
    tet_is_pec[pec_tris] = True

    # CONSTANTS
    N_TETS = tets.shape[1]
    N_EDGES = edges.shape[1]
    N2D = DPTS_2D.shape[1]
    WEIGHTS_VOL = DPTS_3D[0,:]
    
    W0 = k0*C0
    Y0 = np.sqrt(1/MU0)
    
    # INIT POSTERIORI ERROR ESTIMATE QUANTITIES
    alpha_t = np.zeros((N_TETS,), dtype=np.complex128)
    max_elem_size = np.zeros((N_TETS,), dtype=np.float64)
    
    Qf_face1 = np.zeros((4,N2D,N_TETS), dtype=np.complex128)
    Qf_face2 = np.zeros((4,N2D,N_TETS), dtype=np.complex128)
    Jf_face1 = np.zeros((4,3,N2D,N_TETS), dtype=np.complex128)
    Jf_face2 = np.zeros((4,3,N2D,N_TETS), dtype=np.complex128)
    
    areas_face_residual = np.zeros((4, N2D, N_TETS), dtype=np.float64)
    Rf_face_residual = np.zeros((4, N_TETS), dtype=np.float64)
    adj_tets_mat = -np.ones((4,N_TETS), dtype=np.int32)
    
    # Compute Error estimate
    for itet in range(N_TETS):
        uinv = (1/ur[itet])*np.eye(3)
        ermat = er[itet]*np.eye(3)
        erc = er[itet]
        urc = ur[itet]
        
        # GEOMETRIC QUANTITIES
        vertices = nodes[:,tets[:, itet]]
        v1 = vertices[:,0]
        v2 = vertices[:,1]
        v3 = vertices[:,2]
        v4 = vertices[:,3]
        
        # VOLUME INTEGRATION POINTS
        vxs = DPTS_3D[1,:]*v1[0] + DPTS_3D[2,:]*v2[0] + DPTS_3D[3,:]*v3[0] + DPTS_3D[4,:]*v4[0]
        vys = DPTS_3D[1,:]*v1[1] + DPTS_3D[2,:]*v2[1] + DPTS_3D[3,:]*v3[1] + DPTS_3D[4,:]*v4[1]
        vzs = DPTS_3D[1,:]*v1[2] + DPTS_3D[2,:]*v2[2] + DPTS_3D[3,:]*v3[2] + DPTS_3D[4,:]*v4[2]
        
        intpts = np.empty((3,DPTS_3D.shape[1]), dtype=np.float64)
        intpts[0,:] = vxs
        intpts[1,:] = vys
        intpts[2,:] = vzs
        
        # TET TRI NODE COUPLINGS
        g_node_ids = tets[:, itet]
        g_edge_ids = edges[:, tet_to_field[:6, itet]]
        g_tri_ids = tris[:, tet_to_field[6:10, itet]-N_EDGES]
        l_edge_ids = local_mapping(g_node_ids, g_edge_ids)
        l_tri_ids = local_mapping(g_node_ids, g_tri_ids)
        triids = tet_to_tri[:,itet]
        
        size_max = circum_sphere_diam(v1,v2,v3,v4)
        #size_max = np.max(edge_lengths[tet_to_edge[:,itet]])
        
        TET_VOLUME = compute_volume(vertices[0,:], vertices[1,:], vertices[2,:])
        Rt = size_max
        
        # Efield
        Ef = Efield[tet_to_field[:,itet]]
        
        # Qt term
        Qt = TET_VOLUME*EPS0*np.sum(WEIGHTS_VOL*compute_div(intpts, vertices, Ef, l_edge_ids, l_tri_ids, ermat), axis=0)

        # Jt term
        Rv1 = compute_curl_curl(vertices, Ef, l_edge_ids, l_tri_ids, uinv)
        Rv2 = -k0**2*matmul(ermat, compute_field(intpts, vertices, Ef, l_edge_ids, l_tri_ids))
        Rv = 1*Rv2
        Rv[0,:] += Rv1[0] # X-component
        Rv[1,:] += Rv1[1] # Y-component
        Rv[2,:] += Rv1[2] # Z-component
        
        Rv[0,:] = Rv[0,:]*WEIGHTS_VOL
        Rv[1,:] = Rv[1,:]*WEIGHTS_VOL
        Rv[2,:] = Rv[2,:]*WEIGHTS_VOL
        
        Jt = -TET_VOLUME*np.sum(1/(1j*W0*MU0) * Rv, axis=1)
  
        Gt = (1j*W0*np.exp(-1j*k0*Rt)/(4*np.pi*Rt))
        alpha_t[itet] = - Gt/(erc*EPS0) * Qt*Qt - Gt*urc*MU0 * np.sum(Jt*Jt)
        
        # Face Residual computation
        
        all_face_coords = np.empty((3,4*N2D), dtype=np.float64)
        for itri in range(4):
            triid = triids[itri]
            tnodes = nodes[:,tris[:,triid]]
            n1 = tnodes[:,0]
            n2 = tnodes[:,1]
            n3 = tnodes[:,2]
            all_face_coords[0,itri*N2D:(itri+1)*N2D] = DPTS_2D[1,:]*n1[0] + DPTS_2D[2,:]*n2[0] + DPTS_2D[3,:]*n3[0]
            all_face_coords[1,itri*N2D:(itri+1)*N2D] = DPTS_2D[1,:]*n1[1] + DPTS_2D[2,:]*n2[1] + DPTS_2D[3,:]*n3[1]
            all_face_coords[2,itri*N2D:(itri+1)*N2D] = DPTS_2D[1,:]*n1[2] + DPTS_2D[2,:]*n2[2] + DPTS_2D[3,:]*n3[2]
        
        Qf_all = erc*EPS0*compute_field(all_face_coords, vertices, Ef, l_edge_ids, l_tri_ids)
        Jf_all = -1/(1j*MU0*W0)*matmul(uinv, compute_curl(all_face_coords, vertices, Ef, l_edge_ids, l_tri_ids))
        E_face_all = compute_field(all_face_coords, vertices, Ef, l_edge_ids, l_tri_ids)
        tetc = centers[:,itet].flatten()
        
        max_elem_size[itet] = size_max
        
        for iface in range(4):
            tri_index = triids[iface]
            
            pec_face = tet_is_pec[tri_index]
            
            i1, i2, i3 = tris[:, tri_index]
            
            slc1 = iface*N2D
            slc2 = slc1+N2D
            
            normal = outward_normal(nodes[:,i1], nodes[:,i2], nodes[:,i3], tetc).astype(np.complex128)
            
            area = areas[triids[iface]]
            
            n1 = nodes[:,i1]
            n2 = nodes[:,i2]
            n3 = nodes[:,i3]
            l1 = np.linalg.norm(n2-n1)
            l2 = np.linalg.norm(n3-n1)
            l3 = np.linalg.norm(n3-n2)
            Rf = np.max(np.array([l1, l2, l3]))
            Rf = diam_circum_circle(n1,n2,n3)
            Rf_face_residual[iface,itet] = Rf
            areas_face_residual[iface, :, itet] = area
            
            adj_tets = [int(tri_to_tet[j,triids[iface]]) for j in range(2)]
            adj_tets = [num for num in adj_tets if num not in (itet, -1234)]
            
            if len(adj_tets) == 0:
                continue
            
            if pec_face is True:
                Jtan = Y0*np.sqrt(1/urc)*cross_c_arry(normal, -cross_c_arry(normal, E_face_all[:, slc1: slc2]))
                
                itet_adj = adj_tets[0]
                iface_adj = np.argwhere(tet_to_tri[:,itet_adj]==triids[iface])[0][0]
                
                Jf_face1[iface, :, :, itet] = Jtan

                adj_tets_mat[iface,itet] = itet_adj
                continue
            
            itet_adj = adj_tets[0]
            iface_adj = np.argwhere(tet_to_tri[:,itet_adj]==triids[iface])[0][0]
            
            Jf_face1[iface, :, :, itet] = cross_c_arry(normal, Jf_all[:, slc1:slc2])
            Jf_face2[iface_adj, :, :, itet_adj] = -cross_c_arry(normal, Jf_all[:, slc1:slc2])
            
            Qf_face1[iface, :, itet] = dot_c_arry(normal, Qf_all[:, slc1:slc2])
            Qf_face2[iface_adj, :, itet_adj] = -dot_c_arry(normal, Qf_all[:, slc1:slc2])

            adj_tets_mat[iface,itet] = itet_adj

    # Compute 2D Gauss quadrature weight matrix
    fWs = np.empty_like(areas_face_residual, dtype=np.float64)
    for i in range(N2D):
        fWs[:,i,:] = DPTS_2D[0,i]
    
    # Compute the εE field difference (4, NDPTS, NTET)
    Qf_delta = Qf_face1 - Qf_face2
    Jf_delta = Jf_face1 - Jf_face2
    
    # Perform Gauss-Quadrature integration (4, NTET)
    Qf_int = np.sum(Qf_delta*areas_face_residual*fWs, axis=1)
    Jf_int_x = np.sum(Jf_delta[:,0,:,:]*areas_face_residual*fWs, axis=1)
    Jf_int_y = np.sum(Jf_delta[:,1,:,:]*areas_face_residual*fWs, axis=1)
    Jf_int_z = np.sum(Jf_delta[:,2,:,:]*areas_face_residual*fWs, axis=1)
    
    Gf = (1j*W0*np.exp(-1j*k0*Rf_face_residual)/(4*np.pi*Rf_face_residual))
    alpha_Df = - Gf/(er*EPS0)*(Qf_int*Qf_int) - Gf*(ur*MU0) * (Jf_int_x*Jf_int_x + Jf_int_y*Jf_int_y + Jf_int_z*Jf_int_z)
    
    alpha_Nf = np.zeros((4, N_TETS), dtype=np.complex128)
    for it in range(N_TETS):
        for iface in range(4):
            it2 = adj_tets_mat[iface, it]
            if it2==-1:
                continue
            alpha_Nf[iface,it] = alpha_t[it2]
    
    alpha_f = np.sum((alpha_t/(alpha_t + alpha_Nf + 1e-21))*alpha_Df, axis=0)
    
    error = (np.abs(alpha_t + alpha_f))**0.5
    
    return error, max_elem_size

def compute_error_estimate(field: MWField, pec_tris: list[int]) -> tuple[np.ndarray, np.ndarray]:
    """Top level function to compute the EM error of a field solution.

    Args:
        field (MWField): The MWField object to analyse
        pec_tris (list[int]): A list of triangles that ought to be considered PEC (non-neighbouring.)

    Returns:
        np.ndarray, np.ndarray: The error estimate in an (Ntet,) float array and the tetrahedral size value.
    """
    mesh = field.mesh

    nodes = mesh.nodes
    tris = mesh.tris
    tets = mesh.tets
    edges = mesh.edges
    centers = mesh.centers
    
    As = mesh.areas
    tet_to_edge = mesh.tet_to_edge
    tet_to_tri = mesh.tet_to_tri
    tri_centers = mesh.tri_centers
    tri_to_tet = mesh.tri_to_tet
    tet_to_field = field.basis.tet_to_field
    er = field._der
    ur = field._dur
    
    Ls = mesh.edge_lengths
    
    pec_tris = np.sort(np.unique(np.array(pec_tris))).astype(np.int64)
    
    errors = []
    for key in field._fields.keys():
        
        excitation = field._fields[key]
        error, sizes = compute_error_single(nodes, tets, tris, edges,
                             centers, excitation, Ls, As, 
                             tet_to_edge, tet_to_tri, tri_centers,
                             tri_to_tet, tet_to_field, er, ur, pec_tris, field.k0)
        
        errors.append(error)
    
    error = np.max(np.array(errors), axis=0)
    return error, sizes


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

import re
from typing import List, Dict, Optional, Tuple

class ExcellonParseError(Exception):
    pass

def parse_excellon(
    text: str,
    *,
    convert_to: Optional[str] = "mm",   # "mm", "inch", or None for native
) -> List[Dict[str, float]]:
    """
    Parse an Excellon (NC drill) file string and return a list of drill hits with radii.

    Returns: List of dicts with keys:
      - x, y (floats, in 'convert_to' units if provided, else native file units)
      - radius (float, same units)
      - unit ("mm" or "inch")
      - tool (e.g., "T01")

    Features:
      • Supports INCH/METRIC (also M72/M71).
      • Supports zero suppression LZ/TZ.
      • Uses ;FILE_FORMAT=<int>:<dec> if present to interpret implicit decimals.
      • Modal coordinates: missing X or Y reuse last value.
      • Tool table with diameters via lines like: T1F00S00C0.03543
      • Heuristic for “short fields”: If a coordinate has fewer digits than expected,
        it tries both LZ- and TZ-style padding and chooses the value closer to the
        previous axis value (helps with occasional writer quirks).
      • Ignores non-drill routing (basic: lines without X/Y are ignored unless R repeats).
    """
    # Defaults (updated by headers)
    unit = None            # "inch" or "mm"
    zero_sup = "LZ"        # "LZ" or "TZ"
    int_digits, dec_digits = 2, 4  # sensible default if FILE_FORMAT missing
    in_header = False

    # Tooling
    tool_diam_in_native_units: dict[str, float] = {}  # e.g. {"T1": 0.8}
    current_tool: Optional[str] = None

    # Modal state
    last_x: Optional[float] = None
    last_y: Optional[float] = None

    # Results
    hits: List[Dict[str, float]] = []

    # Regex helpers
    re_filefmt = re.compile(r";\s*FILE_FORMAT\s*=\s*(\d+)\s*:\s*(\d+)", re.I)
    re_units_line = re.compile(r"^\s*(INCH|METRIC)\s*(?:,?\s*(LZ|TZ))?\s*$", re.I)
    re_tool_def = re.compile(r"^\s*T(\d+)[^;\r\n]*?C\s*=?\s*([+-]?(?:\d+(?:\.\d*)?|\.\d+))",re.I)
    re_tool_sel = re.compile(r"^\s*T(\d+)\s*$", re.I)
    re_M71 = re.compile(r"^\s*M71\b")  # metric
    re_M72 = re.compile(r"^\s*M72\b")  # inch

    # Coordinate line patterns (allow just X, just Y, or both, and optional R repeat)
    re_xy = re.compile(
        r"^(?:.*?)(?:X(-?\d+))?(?:Y(-?\d+))?(?:R(\d+))?\s*$",
        re.I
    )

    def to_native_units(val: float) -> float:
        return val  # already native; conversion applied later in one go

    def convert(val: float, from_unit: str, to_unit: Optional[str]) -> float:
        if to_unit is None or to_unit == from_unit:
            return val
        if from_unit == "inch" and to_unit == "mm":
            return val * 25.4
        if from_unit == "mm" and to_unit == "inch":
            return val / 25.4
        raise ExcellonParseError(f"Unknown conversion {from_unit}→{to_unit}")

    def parse_number_field(
        s: str,
        last_val: Optional[float],
        int_d: int,
        dec_d: int,
        zero_rule: str
    ) -> float:
        """
        Parse a digit string without a decimal point using the given format.
        Applies LZ/TZ rule, but also tries a heuristic "other side" padding
        and picks the value closer to the last value on that axis if available.
        """
        # If the writer included a decimal point (rare in coords), just float it
        if "." in s or s.startswith(("+", "-")) and "." in s[1:]:
            try:
                return float(s)
            except ValueError:
                raise ExcellonParseError(f"Bad numeric field: {s}")

        total = int_d + dec_d
        raw = s.strip()
        neg = raw.startswith("-")
        if neg:
            raw = raw[1:]
        if not raw.isdigit():
            raise ExcellonParseError(f"Bad digits in coordinate: {s}")

        def to_val(padded: str) -> float:
            # split from the RIGHT by dec_d (more robust if length != expected)
            if dec_d == 0:
                integ, frac = padded, ""
            else:
                frac = padded[-dec_d:] if len(padded) >= dec_d else padded.rjust(dec_d, "0")
                integ = padded[:-dec_d] if len(padded) >= dec_d else ""
            v = (int(integ) if integ else 0) + (int(frac) / (10 ** dec_d) if dec_d > 0 else 0.0)
            return -v if neg else v

        # Primary padding based on declared suppression
        if zero_rule.upper() == "LZ":
            primary = raw.rjust(total, "0")
            alt = raw.ljust(total, "0")
        else:  # TZ
            primary = raw.ljust(total, "0")
            alt = raw.rjust(total, "0")

        v_primary = to_val(primary)
        if last_val is None:
            # no history—trust header rule
            return v_primary

        v_alt = to_val(alt)

        # Choose whichever is closer to the previous modal value on this axis.
        # Require a meaningful improvement to avoid flip-flopping on small noise.
        d_p = abs(v_primary - last_val)
        d_a = abs(v_alt - last_val)

        if d_a + 1e-12 < d_p * 0.2:  # alt is at least 5x closer → take alt
            return v_alt
        return v_primary

    # Pass 1: parse header + body
    for raw_line in text.splitlines():
        line = raw_line.strip()

        # Skip blank lines & pure comments
        if not line or line.startswith(";"):
            # However, parse FILE_FORMAT even if comment-style
            mfmt = re_filefmt.search(raw_line)
            if mfmt:
                int_digits, dec_digits = int(mfmt.group(1)), int(mfmt.group(2))
            continue

        if line.upper().startswith("M48"):
            in_header = True
            continue
        if line == "%":
            # end of header or section; continue
            in_header = False
            continue

        # Units (either in header line or body)
        m_units = re_units_line.match(line)
        if m_units:
            unit = m_units.group(1).lower()
            if m_units.group(2):
                zero_sup = m_units.group(2).upper()
            continue

        if re_M71.match(line):
            unit = "mm"
            continue
        if re_M72.match(line):
            unit = "inch"
            continue

        # Tool definition lines (e.g., T1F00S00C0.03543)
        m_tool_def = re_tool_def.match(line)
        if m_tool_def:
            tnum = int(m_tool_def.group(1))
            tool = f"T{tnum:02d}"
            diam = float(m_tool_def.group(2))
            tool_diam_in_native_units[tool] = diam
            continue

        # Tool selection (e.g., T01)
        m_tool_sel = re_tool_sel.match(line)
        if m_tool_sel:
            tnum = int(m_tool_sel.group(1))
            current_tool = f"T{tnum:02d}"
            # Some writers chain selection with coords on same line, but our regex below catches coords too.
            continue

        # Stop codes etc.
        up = line.upper()
        if up.startswith("M30") or up.startswith("M00") or up.startswith("M95"):
            break

        # Coordinate line: may contain X, Y, optional R
        m_xy = re_xy.match(line)
        if not m_xy:
            # Unknown line; ignore quietly (robust)
            continue

        sx, sy, sr = m_xy.group(1), m_xy.group(2), m_xy.group(3)
        if sx is None and sy is None:
            # nothing here
            continue

        # Parse X/Y with modal reuse if missing
        if sx is not None:
            x_native = parse_number_field(sx, last_x, int_digits, dec_digits, zero_sup)
            last_x = to_native_units(x_native)
        if sy is not None:
            y_native = parse_number_field(sy, last_y, int_digits, dec_digits, zero_sup)
            last_y = to_native_units(y_native)

        if last_x is None or last_y is None:
            # need both to place a hole
            continue

        repeat = int(sr) if sr else 1

        # Determine radius from current tool
        if current_tool and current_tool in tool_diam_in_native_units:
            diam_native = tool_diam_in_native_units[current_tool]
        else:
            # Unknown tool diameter → leave radius as None
            diam_native = None

        # Output in desired units
        final_unit = unit or "mm"  # default to inch if totally unspecified (rare)
        for _ in range(repeat):
            if diam_native is None:
                radius_out = None
            else:
                radius_out = convert(diam_native / 2.0, final_unit, convert_to)
            hits.append({
                "x": convert(last_x, final_unit, convert_to),
                "y": convert(last_y, final_unit, convert_to),
                "radius": radius_out,
                "unit": convert_to or final_unit,
                "tool": current_tool or "T??",
            })

    return hits


def parse_excellon_file(path: str, **kwargs) -> List[Dict[str, float]]:
    with open(path, "r", encoding="utf-8") as f:
        return parse_excellon(f.read(), **kwargs)



import emerge as em
import numpy as np
from emerge.plot import plot_sp, plot_ff_polar, smith

from emerge import EMergeTheme
""" SLOTTED VIVALDI ANTENNA

In this demo we will look at the design of a slotted Vivaldi antenna in EMerge and how it can be setup.

The antenna is not perfect. The aim of this tutorial is to demonstrate how parametric curves are implemented in EMerge.
Additionally, the narrow regions are a perfect opportunity to demonstrate the efficiency of adaptive mesh refinement.

Due to the wind band nature of this simulation, it might take a while to simulate. 

For ARM MacOS users it is reccommended to install UMFPACK and run this using multi-processing.

"""

m = em.Simulation('Vivaldi')
m.check_version("2.1.0")

mm = 0.001          # Millimeter
g = 0.3*mm          # Narrow exponential taper slot gap size
L = 70*mm           # Slot length
W = 55*mm           # Aperturee width
K = 200             # Exponential curve growth factor
Radius = 20*mm      # Open circuit circle radius
th = 0.5            # PCB Thickness
Lstub = 7           # Length of the radial stub
stub_ang = 80       # Radial stub angle
stub_ang_off = 20   # Radial stub angle offset
slot_margin = 5*mm  # Margin around the taper for the slots
wslot = 4*mm        # The width of the slot
wgap = 4*mm         # The distance betwen slots

Wperiod = wslot+wgap


# Vivaldi curve parametric equations fy(t) and fx(t)
# t ∈ [0, 1]
# fx(t) = L*t
# fy(t) = (g/2) Kᵗ + (W-gK)/(2-2K) (1 - Kᵗ)

fx = lambda t: t*L
fy = lambda t: (g/2)*K**t + (W-g*K)/(2-2*K) *(1-K**t)

# We will create the slots around the vivaldi by making slots that stretch across the entire
# PCB and then removing the exponential taper region dilated by some 5 millimeters.
# We could use a curve dilation but in this case we will derive the analytical expresion ourselves!

# This A-coeff is just to make the equations more brief.
A = (g/2) - (W-g*K)/(2-2*K)

# We need to find the derivative of dfy(t)/dfx
dfx = lambda t: A*np.log(K)/L * K**(t)

# To create the dilated curve we have to formulate the parametric equation of a unit length tangent vector
# We know that the ratio of ty/tx is equal to the derivative of the curve so we can normalize the length
# by defining R(t) as
R = lambda t: 1/np.sqrt(1 + dfx(t)**2)
# Such that 
# t = R [f'(x), 1]

# Then the normal vector is simply
# n = R [1, -f'(x)]

# We add this scaled by the margin to our parametric equation to get:
fx2 = lambda t: t*L - slot_margin*R(t)*dfx(t)
fy2 = lambda t: fy(t) + slot_margin*R(t)

# Now that we have our parametric equations defined we first make an airbox and set its material priority to a low value.
# This makes sure that in overlapping regions, the air material will not be assigned. In this example it likely isn't necessary but it is good practice.
airbox = em.geo.Box(120*mm, 80*mm, 30*mm, position=(-30*mm, -40*mm, -15*mm)).prio_set(1)

# Exponential taper
# We make our tapered region as following

# Step 1: We create an XYPolygon that is empty
# Step 2: We draw our first taper for the top curve. That is simply our parametric line
# Step 3: We make the return path being the bottom. This includes the same X-values and the Y-values inverted (-Y). And we must reverse the points from end to start.
# Step 4: Finally we turn our XYPolygon into a geometry by embedding it in a Coordinate System that is at the z-coordinate of the bottom of our PCB (-th). 

exp_taper = em.geo.XYPolygon()\
    .parametric(fx, fy, tolerance=1e-5)\
    .parametric(fx, lambda t: -fy(t), reverse=True, tolerance=1e-5)\
    .geo(em.GCS.displace(0,0,-th*mm))
        
# We do the same for the dilated one
exp_taper_dialated = em.geo.XYPolygon()\
    .parametric(fx2, fy2, tolerance=1e-5, tmax=2)\
    .parametric(fx2, lambda t: -fy2(t), reverse=True, tolerance=1e-5, tmax=2)\
    .geo(em.GCS.displace(0,0,-th*mm))

# Next we generate 15 slots. We make them stretch across the entire PCB.
slots = []
for n in range(7):
    rect = em.geo.XYPlate(wslot, 60*mm, (70*mm - (n+1)*Wperiod, -30*mm, -th*mm))
    slots.append(rect)
    
# Then we unite all the individual slots into one geometry.
slots = em.geo.unite(*slots)
# We subtract the dilated taper from our slots to only have the slots around the taper left.
slots = em.geo.subtract(slots, exp_taper_dialated)
# We also define the circular region that we use to offen an open circuit to the feed point.
disc = em.geo.Disc((-Radius/2+1*mm,0,-th*mm), Radius/2, (0,0,1))

# Our PCB is actually quite simple, it only contains the bottom Ground plane and upper feed trace.
pcbl = em.geo.PCB(th, mm, material=em.lib.DIEL_FR4)
# We compute the trace width for 50 ohms.
w0 = pcbl.calc.z0(50)
# Next we start a new trace at xy=(2mm, -10mm), we store this as our input port and move 10.5mm forwards just beyond where the taper slot is.
pcbl.new(2, -10, w0, (0,1)).store('pin').straight(10.5).store('stub')

# we compute the direction in which our radial stub has to point. We rotate it away from the circle a bit.
dx = np.sin(stub_ang_off*np.pi/180)
dy = np.cos(stub_ang_off*np.pi/180)
rx, ry = pcbl.load('stub').xy # Here we take the X,Y coordinates at the end of our feed line

# Finally we place our stub. We lower is by 0.2mm to make it sit flush with the line due to the rotation.
pcbl.radial_stub((rx, ry-0.2), Lstub, stub_ang, (dx,dy), w0=w0)

# Finallly we compile all the polygons into a single metal trace.
polies = pcbl.compile_paths(merge=True)

# We also create a rectangular region for our lumped port.
port = pcbl.lumped_port(pcbl.load('pin'))

# We manually define the X,Y bounds of the PCB because the vivaldi taper is not part of our PCB itself. 
pcbl.set_bounds(xmin=-25, xmax = 70, ymin=-30, ymax = 30)
pcb = pcbl.generate_pcb() # we generate the PCB delectricum
ground = pcbl.plane(-th) # And ground

# and view the entire geometry
m.view()

# We subtract the exponential taper, disc and slots from the ground plane.
ground = em.geo.subtract(ground, exp_taper)
ground = em.geo.subtract(ground, disc)
ground = em.geo.subtract(ground, slots)

# And we are done modelling!
m.commit_geometry()

# We set our frequency range from 3GHz to 10GHz in 21 setps.
m.mw.set_frequency_range(3e9, 8e9, 15)
m.mw.set_resolution(0.33)

# Here we set our boundary conditions. The Absorbing boundary surfaace is the outside of the airbox.
abc = airbox.outside()

# The lumped port is defined on the port rectangular region. All the dimensions are automatically stored inside
# the port geometry object when you create it with the .lumped_port() function so you don't have to pass them!
m.mw.bc.LumpedPort(port, 1, Z0=50)
m.mw.bc.AbsorbingBoundary(abc)
m.mesher.set_pec_face(polies)
m.mesher.set_pec_face(ground)

m.generate_mesh()

m.view(plot_mesh=True, volume_mesh=False)

# Before we run we call our adaptive mesh refinement at 7GHz. You can change the frequency yourself.
m.adaptive_mesh_refinement(frequency=6e9)
m.view(plot_mesh=True, volume_mesh=False) # and view the resultant mesh

# Finally we start our sweep
data = m.mw.run_sweep(False)

# We extract the object that contains all our gritted global parameter simulation data such as the S-parameters
g = data.scalar.grid

# We create a denser grid of frequency points. This function takes the same value range as the simulation frequency but subsamples
# with a total of 1001 points.
fd = g.dense_f(1001)

# We can plot the S11 in the smith plot. We use model_S(1,1,fd) to interpolate our S-parameters using vector fitting
smith(g.model_S(1,1,fd), f=fd)
S_non_reflected = np.sqrt(1-np.abs(g.model_S(1,1,fd))**2) # The non reflected energy
plot_sp(fd, [g.model_S(1,1,fd), S_non_reflected], labels=['S₁₁','√(1-|S₁₁|²)'])

# We compute the farfield at 6Ghz
ff_data = data.field.find(freq=6e9).farfield_2d((1,0,0),(0,0,1), abc)
plot_ff_polar(ff_data.ang, ff_data.normE/em.EISO, dB=True, title='Farfield E-plane @ 6GHz')

# Finally we create a simple field plot.
m.display.add_objects(*m.all_geos())
m.display.add_field(data.field[1].cutplane(0.8*mm, z=-th*mm/2).scalar('Ey','real'), symmetrize=True)
m.display.show()



import numpy as np
from .optimized import compute_distances
from itertools import groupby

def signed_area(x: np.ndarray, y: np.ndarray) -> float:
    """Compute the signed area of a polygon defined by vertices (x, y)."""
    return 0.5 * np.sum(x * np.roll(y, -1) - y * np.roll(x, -1))

class Loop:
    """This class represents a 2D Polygon loop defined by its x and y coordinates.
    """
    def __init__(self, xs: np.ndarray, ys: np.ndarray):
        self.x: np.ndarray = np.array(xs)
        self.y: np.ndarray = np.array(ys)
        if self.x[0]==self.x[-1]:
            self.x = self.x[:-1]
            self.y = self.y[:-1]
        self.N: int = self.x.shape[0]
        self.solid: bool = False

    def cleanup(self) -> None:
        """
        Remove consecutive duplicate points where both x[i+1] == x[i] and y[i+1] == y[i].
        Updates self.x, self.y, and self.N.
        """
        if self.N == 0:
            return

        mask = np.ones(self.N, dtype=bool)
        mask[1:] = ~((self.x[1:] == self.x[:-1]) & (self.y[1:] == self.y[:-1]))

        self.x = self.x[mask]
        self.y = self.y[mask]
        self.N = self.x.shape[0]
    
    def split(self, merge_margin: float = 1e-9) -> tuple[list[tuple[list[float],list[float]]], list[tuple[list[float],list[float]]]]:
        """Splits the loop into an Add and Remove list of (x,y) points.

        Args:
            merge_margin (float, optional): The distance below which points are considered the same. Defaults to 1e-9.

        Returns:
            tuple[list[tuple[list[float],list[float]]], list[tuple[list[float],list[float]]]]: A tuple containing two lists:
                - The first list contains loops to be added, each represented as a tuple of (x_points, y_points).
                - The second list contains loops to be removed, each represented as a tuple of (x_points, y_points).
        """
        self.cleanup()
        self.unique = np.ones_like(self.x, dtype=np.int32)
        
        ds = compute_distances(self.x, self.y, 0*self.x)
        
        conn = dict()
        # For each polygon node index
        for i in range(self.N):
            # get a list of each connected node that is not itself.
            dsv = np.concatenate((ds[i,:i],[1.0,],ds[i,i+1:]))
            # Truth map of those that are the same
            connected = dsv < merge_margin
            # Map node id -> list of all nodes that are the same [j, k, l]
            conn[i] = list(np.argwhere(connected).flatten())
            # If at least one node is the same, set all unique flags to 0
            if len(conn[i])>0:
                self.unique[np.argwhere(connected)] = 0
        
        # if there are no similar nodes, just return the loop
        if sum(self.unique)==self.unique.shape[0]:
            return ([(self.x, self.y),], [])

        
        id_loop = np.zeros_like(self.x)
        for i, c in conn.items():
            if len(c)!=0:
                id_loop[i]=-1
        
        loopid = 1
        loops = []
        while True:
            if np.all(id_loop != 0):
                break

            start_id = int(np.where(id_loop==0)[0][0])
            id_loop[start_id] = loopid
            current_loop = [start_id,]
            pause_till = -1
            for i in range(start_id+1, self.N):
                if id_loop[i]>0:
                    continue
                if i < pause_till:
                    continue
                if i == pause_till:
                    pause_till = -1
                    continue
                
                if len(conn[i])==0:
                    current_loop.append(i)
                    id_loop[i] = loopid
                else:
                    connected_to = conn[i]
                    if connected_to[0] == current_loop[0]-1:
                        current_loop.append(i)
                        id_loop[i] = loopid
                        break
                    largest = connected_to[-1]
                    current_loop.append(i)
                    id_loop[i] = loopid
                    pause_till = largest
            loopid += 1
            loops.append(current_loop)
        
        areas = [signed_area(self.x[loop], self.y[loop]) for loop in loops]
        np_areas = np.array(areas)
        
        add_loops = []
        remove_loops = []
        
        sign_biggest = np.sign(np_areas[0])
        
        for (loop, A) in zip(loops, areas):
            
            if np.sign(A)*sign_biggest > 0:
                add_loops.append(loop)
            else:
                remove_loops.append(loop)
        
        output_add = [(self.x[loop], self.y[loop]) for loop in add_loops]
        output_remove = [(self.x[loop], self.y[loop]) for loop in remove_loops]
        
        return output_add, output_remove


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

# Last Cleanup: 2026-01-04

from __future__ import annotations
import numpy as np
from typing import Callable

def gauss3_composite(x: np.ndarray, y: np.ndarray) -> float:
    """
    Composite 3-point Gauss (order-2) integral
    on *equally spaced* 1-D data.

    Sections:
        (x0,x1,x2), (x2,x3,x4), (x4,x5,x6), ...

    The rule on one section [x_i, x_{i+2}] is

        ∫ f dx  ≈  (h/2) Σ_k w_k · f( x̄ + (h/2) ξ_k )

    where  h = x_{i+2}−x_i  and
           ξ = (−√3/5, 0, +√3/5),
           w = (8/9, 5/9, 8/9)

    """
    x = np.asarray(x, dtype=np.complex128)
    y = np.asarray(y, dtype=np.complex128)

    if x.ndim != 1 or y.ndim != 1 or x.size != y.size:
        raise ValueError("x and y must be 1-D arrays of equal length.")
    if x.size % 2 == 0:
        raise ValueError("Number of samples must be odd (… 0,1,2; 2,3,4; …).")

    xi = np.sqrt(3/5)
    nodes   = np.array([-xi, 0.0, +xi])
    weights = np.array([5/9, 8/9, 5/9])

    total = 0.0
    for i in range(0, x.size - 2, 2):
        y0, y1, y2 = y[i:i+3]
        a = y0*0.5 - y1 + 0.5*y2
        b = -y0*0.5 + 0.5*y2
        c = y1
        poly_vals = a*nodes**2 + b*nodes + c
        total += np.dot(weights, poly_vals)
    return total

class Line:
    """ A Line class used for convenient definition of integration lines"""
    def __init__(self, xpts: np.ndarray,
                 ypts: np.ndarray,
                 zpts: np.ndarray):
        
        self.xs: np.ndarray = xpts
        self.ys: np.ndarray = ypts
        self.zs: np.ndarray = zpts
        self.dxs: np.ndarray = xpts[1:] - xpts[:-1]
        self.dys: np.ndarray = ypts[1:] - ypts[:-1]
        self.dzs: np.ndarray = zpts[1:] - zpts[:-1]
        self.dl = np.sqrt(self.dxs**2 + self.dys**2 + self.dzs**2)
        self.length: float = np.sum(np.sqrt(self.dxs**2 + self.dys**2 + self.dzs**2))
        self.l: np.ndarray = np.concatenate((np.array([0,]), np.cumsum(self.dl))) 
        self.xmid: np.ndarray = 0.5*(xpts[:-1] + xpts[1:])
        self.ymid: np.ndarray = 0.5*(ypts[:-1] + ypts[1:])
        self.zmid: np.ndarray = 0.5*(zpts[:-1] + zpts[1:])

        self.dx: float = self.dxs[0]
        self.dy: float = self.dys[0]
        self.dz: float = self.dzs[0]
    
    @property
    def cmid(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """The midpoints of the line segments.

        Returns:
            tuple[np.ndarray, np.ndarray, np.ndarray]: The midpoint coordinates (xmid, ymid, zmid).
        """
        return self.xmid, self.ymid, self.zmid

    @property
    def cpoint(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """The center points of the line segments."""
        return self.xs, self.ys, self.zs
    
    @staticmethod
    def from_points(start: np.ndarray, end: np.ndarray, Npts: int) -> Line:
        """Create a Line object from start to end with Npts points.

        Args:
            start (np.ndarray): start point coordinates (x,y,z).
            end (np.ndarray): end point coordinates (x,y,z).
            Npts (int): Number of points along the line.

        Returns:
            Line: The created Line object.
        """
        x1, y1, z1 = start
        x2, y2, z2 = end
        xs = np.linspace(x1, x2, Npts)
        ys = np.linspace(y1, y2, Npts)
        zs = np.linspace(z1, z2, Npts)
        return Line(xs, ys, zs)
    
    @staticmethod
    def from_path(*points: tuple[float,float,float] | np.ndarray, ds: float) -> Line:
        """Creates a Line object from a series of points with approximate spacing ds.

        Args:
            ds (float): Approximate spacing between points.

        Returns:
            Line: The created Line object.
        """
        xpts = []
        ypts = []
        zpts = []
        points = [np.array(p) for p in points]
        xl = None
        yl = None
        zl = None
        for p1, p2 in zip(points[:-1], points[1:]):
            N = max(3,int(np.ceil(np.linalg.norm(p2-p1)/ds)))
            *xs, xl = list(np.linspace(p1[0], p2[0], N))
            *ys, yl = list(np.linspace(p1[1], p2[1], N))
            *zs, zl = list(np.linspace(p1[2], p2[2], N))
            xpts.extend(xs)
            ypts.extend(ys)
            zpts.extend(zs)
        xpts.append(xl)
        ypts.append(yl)
        zpts.append(zl)
        return Line(np.array(xpts), np.array(ypts), np.array(zpts))
            
    def line_integral(self, evalfunc: Callable) -> complex:
        """Compute the line integral for a complex vector field function evalfunc."""
        Ex, Ey, Ez = evalfunc(*self.cpoint)
        EdotL = Ex*self.dx + Ey*self.dy + Ez*self.dz
        return gauss3_composite(self.l, EdotL)
    
    def _integrate(self, quantity: np.ndarray) -> complex:
        """Integrates a quantity of values defined along the line.

        Args:
            quantity (np.ndarray): The quantity values along the line.

        Returns:
            complex: The integrated value.
        """
        return gauss3_composite(self.l, quantity)


import os
import sys
from typing import Tuple, Optional
from .logsettings import DEBUG_COLLECTOR

def _is_microsoft_store_python() -> Tuple[bool, Optional[str]]:
    """
    Detect whether the running interpreter is the Microsoft Store (AppX) build.

    Returns
    -------
    (is_store, evidence)
        is_store : bool
            True if this looks like a Microsoft Store Python.
        evidence : str | None
            Short string explaining what matched, or None if not store.
    """
    if os.name != "nt":
        return False, None

    exe = (sys.executable or "").lower()
    prefix = (sys.prefix or "").lower()
    base_prefix = (getattr(sys, "base_prefix", "") or "").lower()
    exec_prefix = (sys.exec_prefix or "").lower()

    markers = [
        r"\windowsapps\pythonsoftwarefoundation.python.",  # typical Store location
        r"\appdata\local\packages\pythonsoftwarefoundation.python.",  # Store package data
    ]

    for m in markers:
        if m in exe:
            return True, f"sys.executable contains '{m}'"
        if m in prefix:
            return True, f"sys.prefix contains '{m}'"
        if m in base_prefix:
            return True, f"sys.base_prefix contains '{m}'"
        if m in exec_prefix:
            return True, f"sys.exec_prefix contains '{m}'"

    # Fallback: check sys.path entries for Store package layout
    for p in map(lambda s: (s or "").lower(), sys.path):
        if any(m in p for m in markers):
            return True, "a sys.path entry matches Microsoft Store layout"

    return False, None


def run_installation_checks():
    
    # Check 1.
    # Checks if the python installation is from the Microsoft App Store. This will lead to
    # Instabilities with Numba as it can't properly cache in the provided directories.
    
    is_store, why = _is_microsoft_store_python()
    if is_store:
        DEBUG_COLLECTOR.add_report('A Microsoft Store Python installation is detected. This may cause errors when caching Numba code. \n' +
                                'If you run into issues, consider using Conda or UV to install Python on Windows.')



# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from ..._emerge.geo import PCB, extrude
from ..._emerge.geometry import GeoSurface
from .gerber import GerberLayer
from .excellon import parse_excellon_file
from collections import defaultdict

class FileBasedPCB(PCB):
    """Adds PCB construction from Gerber and Excellon files.

    Args:
        PCB (_type_): _description_
    """
    def layer_from_file(self, layer: int, filename: str, res_mm: float = 0.01, n_circ_segments: int = 8, segment_size_mm: float | None = None) -> GeoSurface:
        """Create a layer from a Gerber file.

        Args:
            layer (int): The layer number.
            filename (str): The path to the Gerber file.
            res_mm (float, optional): The resolution in mm. Defaults to 0.01.
            n_circ_segments (int, optional): The number of circular segments. Defaults to 8.
            segment_size_mm (float | None, optional): The size of the segments in mm. Defaults to None.

        Returns:
            GeoSurface: The generated surface.
        """
        gerber = GerberLayer(filename, 0.01, 8, segment_size_mm, cs=self.cs, ignore_via_pads=True)
        z = self.z(layer)*self.unit
        surf = gerber.flatten(z)
        
        self.xs.extend([x/self.unit for x in gerber.xs])
        self.ys.extend([y/self.unit for y in gerber.ys])
        
        if self._thick_traces:
            dx, dy, dz = self.cs.zax.np*self.unit
            vol = extrude(surf, dx, dy, dz).set_material(self.trace_material)
            return vol
        else:
            surf.set_material(self.trace_material)
        return surf
        
    def vias_from_file(self, filename: str, z1: float | None = None, z2: float | None = None, Nsections: int = 6) -> None:
        """Add vias from an Excellon file.

        Args:
            filename (str): The path to the Excellon file.
            z1 (float | None, optional): The starting z-coordinate. Defaults to None.
            z2 (float | None, optional): The ending z-coordinate. Defaults to None.
            Nsections (int, optional): The number of sections for the vias. Defaults to 6.
        """
        vias = []
        via_buffer = defaultdict(list)
        
        for via in parse_excellon_file(filename):
            rad = via['radius']*0.001/self.unit
            x = via['x']*0.001/self.unit
            y = via['y']*0.001/self.unit
            via_buffer[rad].append((x,y))
        
        for size, vias in via_buffer.items():
            self.add_vias(*vias, radius=size, z1=z1, z2=z2, segments=Nsections)
        


# -----------------------------------------------------------------------------
# STEP File import: Dielectric Rod Antenna
#
# This example demonstrates how to import a STEP geometry, define materials,
# assign ports and absorbing boundaries, and run a sweep with far-field output.
# Toggle the if-statements to inspect geometry and mesh behavior.
# -----------------------------------------------------------------------------

import emerge as em
from emerge.plot import plot_sp, plot_ff, plot_ff_polar
from pathlib import Path

# Unit helper: 1 mm in meters
mm = 0.001
Lrod = 70*mm

# Create simulation container
sim = em.Simulation('StepImport')
sim.check_version("2.1.0")

# Set field resolution and frequency sweep
sim.mw.set_resolution(0.25)
sim.mw.set_frequency_range(8e9, 10e9, 21)

# Define custom dielectric material for the rod. We choose something similar to 3D printing resin.
RESIN = em.Material(er=2.5, tand=0.00, color="#777777", opacity=1.0)
# Load STEP file.
# STEPItems groups imported solids. Depending on the configuration of the STEP files, sometimes you have to provide unit=0.001 if the model
# is made in millimeters for example. This does not always go automatically yet.
step_path = Path(__file__).parent / 'DielectricRod.step'
step = em.geo.STEPItems('DielRod', str(step_path), unit=1)

# Unpack labeled volumes.
# Order depends on STEP file: adjust if labels differ in your geometry.
flange, air_cutout, dielectric_rod = step.volumes

# Create an enclosing air region with margin around the geometry.
air = step.enclose(0.01)

# Optional: 
# The order of geometries might not always be evident. To visualize them, create a mesh
# and set labels=True to see approximately which geometry is which which label. The same order
# of numbering is used as as contained in the volumes property.
if False: # Set this to True if you want to view this step.
    sim.commit_geometry()
    sim.generate_mesh()
    sim.view(labels=True)

# You can show all face names as well. This becomes quite a mess
if False:
    sim.commit_geometry()
    sim.generate_mesh()
    sim.view(face_labels=True)
    
# You can also only show the faces of a single geometry
if False:
    sim.commit_geometry()
    sim.generate_mesh()
    sim.view(selections=air_cutout.all_faces())
      
# Assign materials to metal flange and dielectric rod.
flange.set_material(em.lib.PEC)
dielectric_rod.set_material(RESIN)

# Cut port and rod clearance out of the air region and mark result as background material.
# We need to have the port face be backed by an empty space, not air.
air_fin = em.geo.subtract(air, air_cutout).background()

# Commit final geometry and inspect labeled objects and mesh.
# This will crash if you have the previous if statement turned to True.
sim.commit_geometry()
sim.generate_mesh()
sim.view(labels=True)

# Identify port face. This is the top Z face of the cutout region.
# These faces are automatically named
portface = air_fin.face('+z', tool=air_cutout)

# Collect exterior air faces for absorbing boundary assignment.
abc = air_fin.exterior_faces(air)

# Visualize selected exterior boundary region with reduced opacity.
sim.view(selections=[abc], opacity=0.3)

# Define a rectangular waveguide port on the selected face.
port = sim.mw.bc.RectangularWaveguide(portface, 1)

# Apply absorbing boundary condition to outer air faces.
abc_bc = sim.mw.bc.AbsorbingBoundary(abc, abctype='D')

# Run frequency-domain sweep.
data = sim.mw.run_sweep()

# Access structured S-parameter grid.
glob = data.scalar.grid

# Plot reflection coefficient S11 for the input port.
plot_sp(glob.freq, glob.S(1,1))

# Compute 2D far-field cuts at 9 GHz for two principal planes.
ff1 = data.field.find(freq=9e9)\
    .farfield_2d((0, 0, 1), (1, 0, 0), abc)
ff2 = data.field.find(freq=9e9)\
    .farfield_2d((0, 0, 1), (0, 1, 0), abc)

# Plot normalized gain patterns in dBi versus theta.
plot_ff(ff1.ang*180/3.1415,
        [ff1.normE/em.lib.EISO, ff2.normE/em.lib.EISO],
        dB=True, ylabel='Gain [dBi]')

# Polar plot of the same far-field cuts with floor at -20 dB.
plot_ff_polar(ff1.ang,
              [ff1.normE/em.lib.EISO, ff2.normE/em.lib.EISO],
              dB=True, dBfloor=-20)

# Compute full 3D far-field pattern on the absorbing boundary surface.
ffdata = data.field.find(freq=9e9).farfield_3d(abc)

# Add geometry for context in final visualization.
sim.display.add_objects(*sim.all_geos(), opacity=0.1)

# Overlay 3D far-field surface and internal field cut for inspection.
sim.display.add_field(ffdata.surfplot('normE', 'abs',
                                      rmax=40*mm,
                                      offset=(0, 0, Lrod+10*mm)))
sim.display.animate().add_field(
    data.field.find(freq=9e9)
     .cutplane(1*mm, x=0)
     .scalar('Ey', 'complex'),
    symmetrize=True,
)

# Launch interactive visualization window.
sim.display.show()
